from _decimal import Decimal
from datetime import date, timedelta
from typing import List, Optional

from NEMO.models import Account, Project, UsageEvent, User
from NEMO.utilities import RecurrenceFrequency, new_model_copy
from dateutil.relativedelta import relativedelta
from django.test import TestCase
from django.urls import reverse
from django.utils import timezone

from NEMO_billing.admin import save_or_delete_core_facility
from NEMO_billing.cap_discount.customization import CAPDiscountCustomization
from NEMO_billing.cap_discount.exceptions import MissingCAPAmountException
from NEMO_billing.cap_discount.models import CAPDiscount, CAPDiscountAmount, CAPDiscountConfiguration, CAPDiscountTier
from NEMO_billing.cap_discount.processors import AccountDiscountCalculator
from NEMO_billing.invoices.models import BillableItemType, Invoice, InvoiceConfiguration, ProjectBillingDetails
from NEMO_billing.invoices.processors import invoice_data_processor_class as processor
from NEMO_billing.models import CoreFacility
from NEMO_billing.rates.models import Rate, RateCategory, RateType
from NEMO_billing.tests.cap.test_cap_utilities import add_charges_of_type, create_items
from NEMO_billing.tests.test_utilities import basic_data


class TestCapInvoicing(TestCase):
    def setUp(self) -> None:
        self.modify_settings(INSTALLED_APPS={"prepend": "NEMO_billing.cap_discount"})
        self.user, self.project, self.tool, self.area = basic_data(set_rate_category=True)
        cap_configuration = CAPDiscountConfiguration(rate_category=self.project.projectbillingdetails.category)
        # Yearly cap
        cap_configuration.reset_frequency = RecurrenceFrequency.YEARLY.value
        # Started 10 month ago
        cap_configuration.start = date.today() - relativedelta(months=10)
        cap_configuration.start_on_first_charge = False
        # Only for tool usage
        cap_configuration.charge_types = BillableItemType.TOOL_USAGE.value
        cap_configuration.save()
        # 25% discount
        CAPDiscountTier.objects.create(
            cap_discount_configuration=cap_configuration, amount=Decimal("100"), discount=Decimal("25")
        )
        self.cap_configuration: CAPDiscountConfiguration = CAPDiscountConfiguration.objects.get(pk=cap_configuration.pk)
        self.tool, self.area, self.consumable, self.staff = create_items()

    def test_invoice_different_account_cap(self):
        # let's create a new account and generate invoice. nothing should change
        self.cap_configuration.start = None
        self.cap_configuration.save()
        new_project = Project.objects.create(name="new proj", account=Account.objects.create(name="new_acc"))
        new_project.projectbillingdetails = ProjectBillingDetails.objects.create(project=new_project)
        last_month_date = timezone.now() - relativedelta(months=1)
        add_charges_of_type(self, [BillableItemType.TOOL_USAGE], last_month_date)
        self.generate_invoice(last_month_date, new_project.account)
        self.assertFalse(CAPDiscountAmount.objects.filter(cap_discount__account=new_project.account).exists())
        self.assertFalse(CAPDiscountAmount.objects.filter(cap_discount__account=self.project.account).exists())

    def test_invoice_missing_amount_cap_error(self):
        # Let's set some charges 4 months ago, the start date is already set at 10 month ago
        four_month_ago_date = timezone.now() - relativedelta(months=4)
        add_charges_of_type(self, [BillableItemType.TOOL_USAGE], four_month_ago_date)
        # Run an invoice now, that's fine, even if the start date was before
        one_month_ago_date = timezone.now() - relativedelta(months=1)
        invoice = self.generate_invoice(one_month_ago_date)
        # No recent charges, so no invoice
        self.assertFalse(invoice)
        # Delete CAP discount so we can try again
        CAPDiscount.objects.all().delete()
        # Generate invoice 4 months ago
        self.generate_invoice(four_month_ago_date)
        # Try now, this should fail because we are missing invoices for the past 2 months ago
        self.assertRaises(MissingCAPAmountException, self.generate_invoice, one_month_ago_date)
        # Try 2 months ago, same
        two_month_ago_date = timezone.now() - relativedelta(months=2)
        self.assertRaises(MissingCAPAmountException, self.generate_invoice, two_month_ago_date)
        # Try 3 months ago, this time it should work
        three_month_ago_date = timezone.now() - relativedelta(months=3)
        self.generate_invoice(three_month_ago_date)
        # Now we can generate the other ones up to today
        self.generate_invoice(two_month_ago_date)
        self.generate_invoice(one_month_ago_date)

    def test_invoice_missing_amount_cap_error_auto_start(self):
        # Try with start date empty.
        self.cap_configuration.start = None
        self.cap_configuration.start_on_first_charge = True
        self.cap_configuration.save()
        self.test_invoice_missing_amount_cap_error()

    def test_new_month_start(self):
        # Let's set some charges 3 months ago, as well as the start date.
        three_month_ago_date = timezone.now() - relativedelta(months=3)
        self.cap_configuration.start = three_month_ago_date
        self.cap_configuration.save()
        add_charges_of_type(self, [BillableItemType.TOOL_USAGE], three_month_ago_date)
        self.generate_invoice(three_month_ago_date, account=self.project.account)
        # Rate is 30 per hour, the previous function added 24 hours usage
        charge = 30 * 24
        cap_amount = CAPDiscountAmount.objects.get(
            cap_discount__configuration=self.cap_configuration,
            cap_discount__account=self.project.account,
            month=three_month_ago_date.month,
            year=three_month_ago_date.year,
        )
        self.assertEqual(cap_amount.end, charge)
        # new month should start with last month's end amount, and end with + new charges
        two_month_ago_date = timezone.now() - relativedelta(months=2)
        add_charges_of_type(self, [BillableItemType.TOOL_USAGE], two_month_ago_date)
        # Add new charge under the same user/account but different project
        tool_end = two_month_ago_date
        tool_start = tool_end - timedelta(hours=2)
        new_tool_charge = 30 * 2
        new_project = Project.objects.create(name="new proj", account=self.project.account)
        category = RateCategory.objects.get(name="Academia")
        new_project.projectbillingdetails = ProjectBillingDetails.objects.create(project=new_project, category=category)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=new_project, start=tool_start, end=tool_end
        )
        self.generate_invoice(two_month_ago_date, account=self.project.account)
        new_cap_amount = CAPDiscountAmount.objects.get(
            cap_discount__configuration=self.cap_configuration,
            cap_discount__account=self.project.account,
            month=two_month_ago_date.month,
            year=two_month_ago_date.year,
        )
        self.assertEqual(new_cap_amount.start, cap_amount.end)
        self.assertEqual(new_cap_amount.end, cap_amount.end + new_tool_charge + charge)

    def test_month_start_different_facility(self):
        # We already have a cap with no facility, let's create another one with a facility. It should have a start amount after generating invoice
        self.cap_configuration.start = date.today()
        self.cap_configuration.save()
        facility = CoreFacility.objects.create(name="facility")
        new_cap_configuration = new_model_copy(self.cap_configuration)
        new_cap_configuration.save()
        new_cap_configuration.core_facilities.set([facility])
        save_or_delete_core_facility(self.tool, facility, "tool")
        add_charges_of_type(self, [BillableItemType.TOOL_USAGE], timezone.now())
        # Rate is 30 per hour, the previous function added 24 hours usage
        charge = 30 * 24
        self.generate_invoice(timezone.now().astimezone(), account=self.project.account)
        cap_amount = CAPDiscountAmount.objects.filter(cap_discount__configuration=self.cap_configuration).first()
        new_cap_amount = CAPDiscountAmount.objects.filter(cap_discount__configuration=new_cap_configuration).first()
        self.assertTrue(new_cap_amount)
        self.assertEqual(new_cap_amount.start, 0)
        self.assertEqual(new_cap_amount.end, charge)
        # New cap is set for different core facility
        self.assertFalse(cap_amount)

    def test_month_start_different_user(self):
        self.cap_configuration.start = date.today()
        self.cap_configuration.save()
        # Charges added are for the user, so they should not be added to staff's CAP
        add_charges_of_type(self, [BillableItemType.TOOL_USAGE], timezone.now())
        self.generate_invoice(timezone.now().astimezone(), account=self.project.account)
        cap_amount = CAPDiscountAmount.objects.filter(cap_discount__user=self.user).first()
        cap_staff_amount = CAPDiscountAmount.objects.filter(cap_discount__user=self.staff).first()
        self.assertTrue(cap_amount)
        self.assertEqual(cap_amount.start, 0)
        self.assertEqual(cap_amount.end, 24 * 30)
        self.assertFalse(cap_staff_amount)

    def test_excluded_tool(self):
        self.cap_configuration.start = date.today()
        self.cap_configuration.save()
        # Add a tool charge
        add_charges_of_type(self, [BillableItemType.TOOL_USAGE], timezone.now())
        invoice = self.generate_invoice(timezone.now().astimezone(), account=self.project.account)[0]
        cap_amount = CAPDiscountAmount.objects.filter(cap_discount__configuration=self.cap_configuration).first()
        charge = 30 * 24
        discount = (charge - self.cap_configuration.capdiscounttier_set.first().amount) * Decimal(0.25)
        # Everything set correctly
        self.assertTrue(cap_amount)
        self.assertEqual(cap_amount.start, 0)
        self.assertEqual(cap_amount.end, charge)
        self.assertEqual(invoice.total_amount, charge - discount)
        # Now reset and exclude the tool
        invoice.delete()
        CAPDiscount.objects.all().delete()
        CAPDiscountCustomization.set("cap_billing_exclude_tools", f"{self.tool.id}")
        invoice = self.generate_invoice(timezone.now().astimezone(), account=self.project.account)[0]
        cap_amount = CAPDiscountAmount.objects.filter(cap_discount__configuration=self.cap_configuration).first()
        # CAP amount doesn't exist, but invoice stayed unchanged
        self.assertFalse(cap_amount)
        self.assertEqual(invoice.total_amount, charge)

    def test_new_month_reset(self):
        # Let's set some charges and set cap renewal to 2 months
        today_datetime = timezone.now().astimezone()
        two_month_ago_date = today_datetime - relativedelta(months=2)
        self.cap_configuration.start = two_month_ago_date.date()
        self.cap_configuration.reset_interval = 2
        self.cap_configuration.reset_frequency = RecurrenceFrequency.MONTHLY.value
        self.cap_configuration.save()
        add_charges_of_type(self, [BillableItemType.TOOL_USAGE], two_month_ago_date)
        self.generate_invoice(two_month_ago_date, account=self.project.account)
        # try last month
        last_month = today_datetime - relativedelta(months=1)
        self.generate_invoice(last_month, account=self.project.account)
        # Rate is 30 per hour, the previous function added 24 hours usage
        charge = 30 * 24
        cap_amount = CAPDiscountAmount.objects.get(
            cap_discount__configuration=self.cap_configuration, month=last_month.month, year=last_month.year
        )
        # Should have start and end amount be the same since we have no new charges
        self.assertEqual(cap_amount.end, charge)
        self.assertEqual(cap_amount.start, charge)
        # Now generate new invoice. cap should reset
        self.generate_invoice(today_datetime, account=self.project.account)
        cap_amount = CAPDiscountAmount.objects.get(
            cap_discount__configuration=self.cap_configuration, month=today_datetime.month, year=today_datetime.year
        )
        self.assertEqual(cap_amount.start, 0)
        self.assertEqual(cap_amount.end, 0)

    def test_rollback_if_one_project_invoice_has_error(self):
        pass

    def test_discount_invoice_same_project(self):
        # Our cap is at 100, and discount 25%. Set start 2 month ago
        two_month_ago_date = timezone.now() - relativedelta(months=2)
        self.cap_configuration.start = two_month_ago_date
        self.cap_configuration.save()
        # Add charges 2 months ago, make sure the end is earlier, so it counts for cap before the others
        end = two_month_ago_date - timedelta(minutes=20)
        start = end - timedelta(hours=2)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=self.project, start=start, end=end
        )
        # 30 per hour for 2 hours
        charge_1 = 30 * 2
        last_invoice: Invoice = self.generate_invoice(two_month_ago_date, self.project.account)[0]
        self.assertEqual(last_invoice.total_amount, charge_1)
        # Reset
        last_invoice.delete()
        CAPDiscount.objects.all().delete()
        # Add new charge for one hour, still no discount
        end = two_month_ago_date - timedelta(minutes=15)
        start = end - timedelta(hours=1)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=self.project, start=start, end=end
        )
        charge_2 = 30 * 1
        last_invoice: Invoice = self.generate_invoice(two_month_ago_date, self.project.account)[0]
        self.assertEqual(last_invoice.total_amount, charge_1 + charge_2)
        # Add new charge for one hour, this time it should be discounted 25% on the difference
        last_invoice.delete()
        CAPDiscount.objects.all().delete()
        end = two_month_ago_date - timedelta(minutes=10)
        start = end - timedelta(hours=1)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=self.project, start=start, end=end
        )
        # 25% discount
        charge_3 = 30 * 1 - (20 * 0.25)
        last_invoice: Invoice = self.generate_invoice(two_month_ago_date, self.project.account)[0]
        self.assertEqual(last_invoice.total_amount, charge_1 + charge_2 + charge_3)
        # Add new charges to different project, but under the same account. It should count towards CAP
        # And the second project should have the discount as well
        last_invoice.delete()
        CAPDiscount.objects.all().delete()
        new_project = Project.objects.create(name="new proj", account=self.project.account)
        category = RateCategory.objects.get(name="Academia")
        new_project.projectbillingdetails = ProjectBillingDetails.objects.create(project=new_project, category=category)
        end = two_month_ago_date - timedelta(minutes=5)
        start = end - timedelta(hours=2)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=new_project, start=start, end=end
        )
        charge_4 = 30 * 2 - (30 * 2 * 0.25)
        invoices = self.generate_invoice(two_month_ago_date, self.project.account)
        self.assertEqual(len(invoices), 2)
        project_invoice: Invoice = [invoice for invoice in invoices if invoice.project_details.project == self.project][
            0
        ]
        self.assertEqual(project_invoice.total_amount, charge_1 + charge_2 + charge_3)
        new_project_invoice: Invoice = [
            invoice for invoice in invoices if invoice.project_details.project == new_project
        ][0]
        self.assertEqual(new_project_invoice.total_amount, charge_4)

    def test_discount_invoice_no_users(self):
        # Set configuration to NOT split by users
        self.cap_configuration.split_by_user = False
        # Our cap is at 100, and discount 25%. Set start 2 month ago
        two_month_ago_date = timezone.now() - relativedelta(months=2)
        self.cap_configuration.start = two_month_ago_date
        self.cap_configuration.save()
        # Add charges 2 months ago, make sure the end is earlier, so it counts for cap before the others
        end = two_month_ago_date - timedelta(minutes=20)
        start = end - timedelta(hours=2)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=self.project, start=start, end=end
        )
        # 30 per hour for 3 hours
        charge_1 = 30 * 2
        last_invoice: Invoice = self.generate_invoice(two_month_ago_date, self.project.account)[0]
        self.assertEqual(last_invoice.total_amount, charge_1)
        # Reset
        last_invoice.delete()
        CAPDiscount.objects.all().delete()
        # Add new charge for one hour, still no discount
        end = two_month_ago_date - timedelta(minutes=15)
        start = end - timedelta(hours=1)
        UsageEvent.objects.create(
            user=self.staff, operator=self.staff, tool=self.tool, project=self.project, start=start, end=end
        )
        charge_2 = 30 * 1
        last_invoice: Invoice = self.generate_invoice(two_month_ago_date, self.project.account)[0]
        self.assertEqual(last_invoice.total_amount, charge_1 + charge_2)
        # Add new charge for one hour, this time it should be discounted 25% on the difference
        last_invoice.delete()
        CAPDiscount.objects.all().delete()
        end = two_month_ago_date - timedelta(minutes=10)
        start = end - timedelta(hours=1)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=self.project, start=start, end=end
        )
        # 25% discount on 20 of the $30 charge (reached 100)
        charge_3 = 30 * 1 - (20 * 0.25)
        last_invoice: Invoice = self.generate_invoice(two_month_ago_date, self.project.account)[0]
        self.assertEqual(last_invoice.total_amount, charge_1 + charge_2 + charge_3)
        # Add new charges to different project, but under the same account. It should count towards CAP
        # And the second project should have the discount as well
        last_invoice.delete()
        CAPDiscount.objects.all().delete()
        new_project = Project.objects.create(name="new proj", account=self.project.account)
        category = RateCategory.objects.get(name="Academia")
        new_project.projectbillingdetails = ProjectBillingDetails.objects.create(project=new_project, category=category)
        end = two_month_ago_date - timedelta(minutes=5)
        start = end - timedelta(hours=2)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=new_project, start=start, end=end
        )
        charge_4 = 30 * 2 - (30 * 2 * 0.25)
        invoices = self.generate_invoice(two_month_ago_date, self.project.account)
        self.assertEqual(len(invoices), 2)
        project_invoice: Invoice = [invoice for invoice in invoices if invoice.project_details.project == self.project][
            0
        ]
        self.assertEqual(project_invoice.total_amount, charge_1 + charge_2 + charge_3)
        new_project_invoice: Invoice = [
            invoice for invoice in invoices if invoice.project_details.project == new_project
        ][0]
        self.assertEqual(new_project_invoice.total_amount, charge_4)
        # Add new discount tier > 140 => 50%
        self.cap_configuration.capdiscounttier_set.add(
            CAPDiscountTier.objects.create(cap_discount_configuration=self.cap_configuration, amount=140, discount=50)
        )
        for invoice in invoices:
            invoice.delete()
        CAPDiscount.objects.all().delete()
        invoices = self.generate_invoice(two_month_ago_date, self.project.account)
        self.assertEqual(len(invoices), 2)
        project_invoice: Invoice = [invoice for invoice in invoices if invoice.project_details.project == self.project][
            0
        ]
        # charges 1-3 are unchanged (total of 120 < 140)
        self.assertEqual(project_invoice.total_amount, charge_1 + charge_2 + charge_3)
        new_project_invoice: Invoice = [
            invoice for invoice in invoices if invoice.project_details.project == new_project
        ][0]
        # charge 4 is now split, 20 to reach 140 at 25%, then the remaining 40 at 50% discount
        charge_4 = 30 * 2 - (20 * 0.25) - (40 * 0.5)
        self.assertEqual(new_project_invoice.total_amount, charge_4)

    def test_void_related_invoices(self):
        accounting = User.objects.create(
            first_name="Staffy", last_name="McStaffer", username="staffy", is_accounting_officer=True
        )
        # Our cap is at 100, and discount 25%. Set start 3 month ago
        three_month_ago_date = timezone.now() - relativedelta(months=3)
        two_month_ago_date = timezone.now() - relativedelta(months=2)
        self.cap_configuration.start = three_month_ago_date
        self.cap_configuration.save()
        # Add charges 3 months ago
        end = three_month_ago_date - timedelta(minutes=20)
        start = end - timedelta(hours=2)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=self.project, start=start, end=end
        )
        # Add charges 2 months ago
        end = two_month_ago_date - timedelta(minutes=20)
        start = end - timedelta(hours=2)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=self.project, start=start, end=end
        )
        # Add new charges to different project, but under the same account.
        new_project = Project.objects.create(name="new proj", account=self.project.account)
        category = RateCategory.objects.get(name="Academia")
        new_project.projectbillingdetails = ProjectBillingDetails.objects.create(project=new_project, category=category)
        end = two_month_ago_date - timedelta(minutes=5)
        start = end - timedelta(hours=2)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=new_project, start=start, end=end
        )
        # Generate invoices 3 months ago, then 2 months ago
        first_invoices = self.generate_invoice(three_month_ago_date, self.project.account)
        invoices = self.generate_invoice(two_month_ago_date, self.project.account)
        self.assertEqual(len(invoices), 2)
        self.client.force_login(accounting)
        # Void one of the earliest invoices, should fail
        response = self.client.post(reverse("void_invoice", args=[first_invoices[0].id]), follow=True)
        self.assertContains(
            response, "You can only void/delete the latest invoices when a CAP is associated with it", status_code=200
        )
        # Void one of the latest invoices, the other one should also be voided and the amounts should be deleted as well
        self.client.post(reverse("void_invoice", args=[invoices[0].id]), follow=True)
        self.assertTrue(
            all(
                [
                    invoice.voided_date and invoice.voided_by == accounting
                    for invoice in Invoice.objects.filter(id__in=[invoice.id for invoice in invoices])
                ]
            )
        )
        self.assertFalse(
            CAPDiscountAmount.objects.filter(
                cap_discount__account=self.project.account, year=two_month_ago_date.year, month=two_month_ago_date.month
            ).exists()
        )
        # CAP discount should still exist
        self.assertTrue(CAPDiscount.objects.filter(account=self.project.account).exists())
        # Now void previous one, we have nothing left
        self.client.post(reverse("void_invoice", args=[first_invoices[0].id]), follow=True)
        self.assertTrue(Invoice.objects.get(pk=first_invoices[0].id).voided_date)
        self.assertFalse(CAPDiscountAmount.objects.all().exists())
        self.assertFalse(CAPDiscount.objects.all().exists())

    def test_delete_related_invoices(self):
        accounting = User.objects.create(
            first_name="Staffy", last_name="McStaffer", username="staffy", is_accounting_officer=True
        )
        # Our cap is at 100, and discount 25%. Set start 3 month ago
        three_month_ago_date = timezone.now() - relativedelta(months=3)
        two_month_ago_date = timezone.now() - relativedelta(months=2)
        self.cap_configuration.start = three_month_ago_date
        self.cap_configuration.save()
        # Add charges 3 months ago
        end = three_month_ago_date - timedelta(minutes=20)
        start = end - timedelta(hours=2)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=self.project, start=start, end=end
        )
        # Add charges 2 months ago
        end = two_month_ago_date - timedelta(minutes=20)
        start = end - timedelta(hours=2)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=self.project, start=start, end=end
        )
        # Add new charges to different project, but under the same account.
        new_project = Project.objects.create(name="new proj", account=self.project.account)
        category = RateCategory.objects.get(name="Academia")
        new_project.projectbillingdetails = ProjectBillingDetails.objects.create(project=new_project, category=category)
        end = two_month_ago_date - timedelta(minutes=5)
        start = end - timedelta(hours=2)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=new_project, start=start, end=end
        )
        # Generate invoices 3 months ago, then 2 months ago
        first_invoices = self.generate_invoice(three_month_ago_date, self.project.account)
        invoices = self.generate_invoice(two_month_ago_date, self.project.account)
        self.assertEqual(len(invoices), 2)
        self.client.force_login(accounting)
        # Delete one of the earliest invoices, should fail
        response = self.client.post(reverse("delete_invoice", args=[first_invoices[0].id]), follow=True)
        self.assertContains(
            response, "You can only void/delete the latest invoices when a CAP is associated with it", status_code=200
        )
        # Delete one of the latest invoices, the other one should also be deleted and the amounts should be deleted as well
        self.client.post(reverse("delete_invoice", args=[invoices[0].id]), follow=True)
        self.assertFalse(Invoice.objects.filter(id__in=[invoice.id for invoice in invoices]).exists())
        self.assertFalse(
            CAPDiscountAmount.objects.filter(
                cap_discount__account=self.project.account, year=two_month_ago_date.year, month=two_month_ago_date.month
            ).exists()
        )
        # CAP discount should still exist
        self.assertTrue(CAPDiscount.objects.filter(account=self.project.account).exists())
        # Now delete previous one, we have nothing left
        self.client.post(reverse("delete_invoice", args=[first_invoices[0].id]), follow=True)
        self.assertRaises(Invoice.DoesNotExist, Invoice.objects.get, pk=first_invoices[0].id)
        self.assertFalse(CAPDiscountAmount.objects.all().exists())
        self.assertFalse(CAPDiscount.objects.all().exists())

    def test_void_invoice_without_cap(self):
        accounting = User.objects.create(
            first_name="Staffy", last_name="McStaffer", username="staffy", is_accounting_officer=True
        )
        # Our cap is at 100, and discount 25%. Set start 2 month ago
        two_month_ago_date = timezone.now() - relativedelta(months=2)
        self.cap_configuration.start = two_month_ago_date
        self.cap_configuration.save()
        # Add charges 2 months ago
        end = two_month_ago_date - timedelta(minutes=20)
        start = end - timedelta(hours=2)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=self.project, start=start, end=end
        )
        # Add new charges to different project, and different account (not capped -> industry).
        new_project = Project.objects.create(name="new proj", account=Account.objects.create(name="Test Account 2"))
        category = RateCategory.objects.get(name="Industry")
        tool_usage = RateType.objects.get(type="TOOL_USAGE")
        Rate.objects.create(category=category, type=tool_usage, tool=self.tool, amount="30.00")
        new_project.projectbillingdetails = ProjectBillingDetails.objects.create(project=new_project, category=category)
        end = two_month_ago_date - timedelta(minutes=5)
        start = end - timedelta(hours=2)
        UsageEvent.objects.create(
            user=self.user, operator=self.user, tool=self.tool, project=new_project, start=start, end=end
        )
        # Generate invoices 2 months ago
        invoices = self.generate_invoice(two_month_ago_date, self.project.account)
        invoices.extend(self.generate_invoice(two_month_ago_date, new_project.account))
        self.assertEqual(len(invoices), 2)
        self.client.force_login(accounting)
        self.client.post(reverse("delete_invoice", args=[invoices[0].id]), follow=True)
        self.assertRaises(Invoice.DoesNotExist, Invoice.objects.get, pk=invoices[0].id)
        self.assertTrue(Invoice.objects.get(pk=invoices[1].id))

    def test_multi_tier_discount(self):
        cap_discount = self.cap_configuration.get_or_create_cap_discount(
            account=self.project.account, username=None, start=timezone.now()
        )
        # CAP is at 25% for > 100
        calculator = AccountDiscountCalculator(timezone.now())
        self.assertEqual(calculator.calculate_discount(50, 50, cap_discount), 0)
        self.assertEqual(calculator.calculate_discount(100, 40, cap_discount), -10)
        self.assertEqual(calculator.calculate_discount(100, 100, cap_discount), -25)
        # Add new tier for over 400 -> 50%
        cap_discount.capdiscounttier_set.add(
            CAPDiscountTier.objects.create(cap_discount=cap_discount, amount=400, discount=50)
        )
        self.assertEqual(calculator.calculate_discount(50, 50, cap_discount), 0)
        self.assertEqual(calculator.calculate_discount(100, 40, cap_discount), -10)
        self.assertEqual(calculator.calculate_discount(100, 100, cap_discount), -25)
        self.assertEqual(calculator.calculate_discount(300, 100, cap_discount), -25)
        # discount -> 100 * 50% (50) + 100 * 25% (25) = -75
        self.assertEqual(calculator.calculate_discount(300, 200, cap_discount), -75)
        # discount -> 200 * 50% (100) + 100 * 25% (25) = -125
        self.assertEqual(calculator.calculate_discount(300, 300, cap_discount), -125)
        # Add new tier for over 400 -> 50%
        cap_discount.capdiscounttier_set.add(
            CAPDiscountTier.objects.create(cap_discount=cap_discount, amount=600, discount=75)
        )
        self.assertEqual(calculator.calculate_discount(300, 300, cap_discount), -125)
        self.assertEqual(calculator.calculate_discount(50, 50, cap_discount), 0)
        self.assertEqual(calculator.calculate_discount(100, 40, cap_discount), -10)
        self.assertEqual(calculator.calculate_discount(100, 100, cap_discount), -25)
        self.assertEqual(calculator.calculate_discount(300, 100, cap_discount), -25)
        self.assertEqual(calculator.calculate_discount(300, 200, cap_discount), -75)
        self.assertEqual(calculator.calculate_discount(300, 300, cap_discount), -125)
        # discount -> 500 * 75% (375) + 200 * 50% (100) + 300 * 25% (75) = -550
        self.assertEqual(calculator.calculate_discount(100, 1000, cap_discount), -550)

    def generate_invoice(self, invoice_date, account: Account = None) -> Optional[List[Invoice]]:
        invoice_configuration = InvoiceConfiguration.first_or_default()
        if not invoice_configuration.id:
            invoice_configuration.save()
        month = invoice_date.strftime("%B %Y")
        account = account or self.project.account
        return processor.generate_invoice_for_account(month, account, invoice_configuration, self.user)



# Generated by Django 4.2.20 on 2025-07-10 22:39

import NEMO.fields
import NEMO_billing.quotes.models
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("NEMO", "0112_fix_project_name_uniqueness"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Quote",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(help_text="Name of the quote", max_length=100)),
                (
                    "emails",
                    NEMO.fields.MultiEmailField(
                        blank=True,
                        help_text="Email addresses to send this quote to. A comma-separated list can be used",
                        max_length=2000,
                        null=True,
                    ),
                ),
                ("expiration_date", models.DateField(blank=True, help_text="Date when this quote expires", null=True)),
                (
                    "created_date",
                    models.DateTimeField(auto_now_add=True, help_text="Date and time when this quote was created"),
                ),
                (
                    "updated_date",
                    models.DateTimeField(auto_now=True, help_text="Date and time when this quote was last updated"),
                ),
                (
                    "published_date",
                    models.DateTimeField(
                        blank=True, help_text="Date and time when this quote was published", null=True
                    ),
                ),
                (
                    "last_emails_sent_date",
                    models.DateTimeField(
                        blank=True, help_text="Date and time when the last emails were sent for this quote", null=True
                    ),
                ),
                (
                    "status",
                    models.IntegerField(
                        choices=[(0, "Draft"), (1, "Pending Approval"), (2, "Published")],
                        default=0,
                        help_text="Current status of the quote",
                    ),
                ),
                (
                    "file",
                    models.FileField(
                        blank=True,
                        max_length=255,
                        null=True,
                        upload_to=NEMO_billing.quotes.models.get_quote_document_filename,
                    ),
                ),
                (
                    "file_access_token",
                    models.CharField(
                        blank=True,
                        help_text="A random token used to access the quote file without authentication",
                        max_length=255,
                        null=True,
                    ),
                ),
                (
                    "approved_by",
                    models.ForeignKey(
                        blank=True,
                        help_text="User who approved or published this quote",
                        null=True,
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="quotes_approved",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="QuoteConfiguration",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "name",
                    models.CharField(help_text="The name of this quote configuration", max_length=255, unique=True),
                ),
                (
                    "expiration_in_days",
                    models.PositiveIntegerField(
                        default=30, help_text="The default number of days quotes are valid for once published"
                    ),
                ),
                ("email_from", models.EmailField(help_text="The email address used to send quotes", max_length=254)),
                (
                    "email_cc",
                    NEMO.fields.MultiEmailField(
                        blank=True,
                        help_text="Email to cc the quote to. A comma-separated list can be used",
                        max_length=2000,
                        null=True,
                    ),
                ),
                ("merchant_name", models.CharField(max_length=255)),
                (
                    "merchant_details",
                    models.TextField(
                        blank=True,
                        help_text="The merchant details to be included in the quote (address, phone number etc.)",
                        null=True,
                    ),
                ),
                ("merchant_logo", models.ImageField(blank=True, null=True, upload_to="quote_logos")),
                (
                    "terms",
                    models.TextField(
                        blank=True, help_text="Terms and conditions to be included in the quote", null=True
                    ),
                ),
                ("currency", models.CharField(default="USD", max_length=4)),
                ("currency_symbol", models.CharField(blank=True, default="$", max_length=4, null=True)),
                (
                    "tax",
                    models.DecimalField(
                        blank=True,
                        decimal_places=3,
                        help_text="Tax in percent. For 20.5% enter 20.5",
                        max_digits=5,
                        null=True,
                    ),
                ),
                ("tax_name", models.CharField(blank=True, default="VAT", max_length=100, null=True)),
                (
                    "create_permissions",
                    NEMO.fields.MultiRoleGroupPermissionChoiceField(
                        help_text="The roles/groups required for users to create this type of quote"
                    ),
                ),
                (
                    "approval_permissions",
                    NEMO.fields.MultiRoleGroupPermissionChoiceField(
                        blank=True,
                        help_text="The roles/groups required for users to approve this type of quote",
                        null=True,
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="QuoteItem",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "quantity",
                    models.DecimalField(decimal_places=0, default=1, help_text="Quantity of the item", max_digits=8),
                ),
                ("amount", models.DecimalField(decimal_places=2, help_text="Amount for this item", max_digits=8)),
                (
                    "minimum_charge",
                    models.DecimalField(
                        blank=True, decimal_places=2, help_text="Minimum amount for this item", max_digits=8, null=True
                    ),
                ),
                ("description", models.CharField(help_text="Description for this item", max_length=255)),
                (
                    "rate_type",
                    models.IntegerField(
                        choices=[(0, "N/A"), (1, "Flat"), (2, "Hourly"), (3, "Daily")],
                        default=0,
                        help_text="Rate type (N/A, Flat, Hourly, Daily) for this item",
                    ),
                ),
                ("quote", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="quotes.quote")),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddField(
            model_name="quote",
            name="configuration",
            field=models.ForeignKey(
                help_text="Configuration for this quote",
                on_delete=django.db.models.deletion.PROTECT,
                to="quotes.quoteconfiguration",
            ),
        ),
        migrations.AddField(
            model_name="quote",
            name="creator",
            field=models.ForeignKey(
                help_text="User who created this quote",
                on_delete=django.db.models.deletion.PROTECT,
                related_name="quotes_created",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="quote",
            name="project",
            field=models.ForeignKey(
                blank=True,
                help_text="Project associated with this quote",
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                to="NEMO.project",
            ),
        ),
        migrations.AddField(
            model_name="quote",
            name="users",
            field=models.ManyToManyField(
                blank=True, help_text="Users associated with this quote", to=settings.AUTH_USER_MODEL
            ),
        ),
    ]



from datetime import datetime
from itertools import groupby
from typing import Iterable, List, Tuple

from NEMO.decorators import accounting_or_manager_required, user_passes_test
from NEMO.models import Area, Consumable, Tool
from NEMO.utilities import distinct_qs_value_list, format_datetime
from NEMO.views.pagination import SortedPaginator
from django.conf import settings
from django.db.models import Q, QuerySet
from django.http import HttpResponse, HttpResponseBadRequest
from django.shortcuts import get_object_or_404, render
from django.utils.html import escape
from django.utils.safestring import mark_safe
from django.views.decorators.http import require_GET, require_http_methods

from NEMO_billing.invoices.utilities import flatten
from NEMO_billing.rates.admin import RateAdminForm
from NEMO_billing.rates.customization import BillingRatesCustomization
from NEMO_billing.rates.models import Rate, RateCategory, RateTime, RateType
from NEMO_billing.rates.utilities import RateHistory

OTHER_RATE_TYPE = "other"


@accounting_or_manager_required
@require_GET
def rates(request, rate_type_choice: str = None):
    rate_type = rate_type_choice or RateType.Type.TOOL
    rate_types = get_rate_types(rate_type)
    rate_order_by = [rate_type.lower(), "type", "category", "time"]
    if rate_type_choice and rate_type_choice.lower() == OTHER_RATE_TYPE:
        rate_types = RateType.objects.filter(item_specific=False)
        rate_order_by = []
    rate_qs = Rate.non_deleted()
    page_qs = rate_qs.filter(type__in=rate_types).order_by(*rate_order_by)
    page = SortedPaginator(page_qs, request, order_by="type" if not rate_order_by else None).get_current_page()
    # remove when fixed in NEMO (allowing no order_type)
    if rate_order_by:
        page.paginator.order_by = ""

    # Creating rates list for search autocomplete
    rate_search = "["
    for tool_id in distinct_qs_value_list(rate_qs.filter(tool__isnull=False), "tool"):
        tool = Tool.objects.get(pk=tool_id)
        rate_search += '{{"name":"{0}", "id":{1}, "type_value": "{2}"}},'.format(
            escape(tool.name), tool_id, RateType.Type.TOOL
        )
    for area_id in distinct_qs_value_list(rate_qs.filter(area__isnull=False), "area"):
        area = Area.objects.get(pk=area_id)
        rate_search += '{{"name":"{0}", "id":{1}, "type_value": "{2}"}},'.format(
            escape(area.name), area_id, RateType.Type.AREA
        )
    for consumable_id in distinct_qs_value_list(rate_qs.filter(consumable__isnull=False), "consumable"):
        consumable = Consumable.objects.get(pk=consumable_id)
        rate_search += '{{"name":"{0}", "id":{1}, "type_value": "{2}"}},'.format(
            escape(consumable.name), consumable_id, RateType.Type.CONSUMABLE
        )
    for rate_type_item in RateType.objects.all():
        if not rate_type_item.item_specific:
            if rate_qs.filter(type=rate_type_item).exists():
                rate_search += '{{"name":"{0}", "id":"{1}", "type_value": "{2}"}},'.format(
                    escape(str(rate_type_item)), "", rate_type_item.get_rate_group_type()
                )
    rate_search = rate_search.rstrip(",") + "]"
    rate_search = mark_safe(rate_search)

    dictionary = {
        "rate_type": rate_type,
        "page": page,
        "show_type": page_qs.values("type").distinct().count() > 1 or rate_type.lower() == OTHER_RATE_TYPE,
        "show_category": page_qs.filter(category__isnull=False).exists(),
        "show_time": page_qs.filter(time__isnull=False).exists(),
        "search_rates": rate_search,
    }

    return render(request, "rates/rates.html", dictionary)


@accounting_or_manager_required
@require_http_methods(["GET", "POST"])
def create_or_modify_rate(request, rate_type_choice=None, item_id=None):
    tool = get_object_or_404(Tool, pk=item_id) if item_id and rate_type_choice == RateType.Type.TOOL else None
    area = get_object_or_404(Area, pk=item_id) if item_id and rate_type_choice == RateType.Type.AREA else None
    consumable = (
        get_object_or_404(Consumable, pk=item_id) if item_id and rate_type_choice == RateType.Type.CONSUMABLE else None
    )

    # All tools, visible or not, should be able to have a rate
    tools = list(Tool.objects.all())
    # Because we are showing tool children, we cannot sort the queryset (child tool has no category)
    tools.sort(key=lambda x: (x.category, x.name))
    rate_types = get_rate_types(rate_type_choice)

    dictionary = {
        "rate_types": rate_types,
        "rate_type_choice": rate_type_choice,
        "item": tool or area or consumable,
        "rate_type_choices": get_rate_type_choices(),
        "rate_categories": RateCategory.objects.all(),
        "rate_times": RateTime.objects.all(),
        "tools": tools,
        "areas": Area.objects.filter(area_children_set__isnull=True),
        "consumables": Consumable.objects.all().order_by("category", "name"),
    }

    if request.method == "GET":
        rate_forms = get_forms_sorted_by_category(request, rate_types, tool, area, consumable)
        dictionary["forms"] = rate_forms
        return render(request, "rates/rate.html", dictionary)
    else:
        rate_form = RateAdminForm(request.POST, rate_history=True)
        if rate_form.is_valid():
            # Don't commit since we want to save it using our own method
            form_rate: Rate = rate_form.save(commit=False)
            form_rate.save_with_user(request.user)
            rate_form.save_m2m()
            # Reinitialize the form now that we know everything went well
            rate_form = init_form(
                request,
                form_rate.type,
                form_rate.flat,
                form_rate.tool,
                form_rate.area,
                form_rate.consumable,
                form_rate.category,
            )
        dictionary["rate_form"] = rate_form
        dictionary["index"] = request.POST.get("index", 0)
        return render(request, "rates/rate_form_details.html", dictionary)


# We are building forms for each rate type, for each category.
# The goal is to have all rate forms related to one group type (Tool, Area, Supply etc.) in one view.
def get_forms_sorted_by_category(
    request, rate_types: QuerySet[RateType], tool, area, consumable
) -> List[RateAdminForm]:
    forms = []
    categories = RateCategory.objects.all()
    missed_reservation_flat = getattr(settings, "DEFAULT_MISSED_RESERVATION_FLAT", True)
    # Create a form for each rate type
    for rate_type in rate_types:
        missed_r_type = rate_type.type in [RateType.Type.TOOL_MISSED_RESERVATION, RateType.Type.AREA_MISSED_RESERVATION]
        default_flat = missed_r_type and missed_reservation_flat or rate_type.type == RateType.Type.CONSUMABLE
        if rate_type.category_specific and categories:
            # Add a form for each category
            for category in categories:
                forms.append(init_form(request, rate_type, default_flat, tool, area, consumable, category))
        else:
            forms.append(init_form(request, rate_type, default_flat, tool, area, consumable))
    forms.sort(key=lambda form: (form.instance.category.name if form.instance.category else ""))
    return forms


def init_form(request, rate_type, flat, tool=None, area=None, consumable=None, category=None) -> RateAdminForm:
    instance = Rate(type=rate_type, category=category, tool=tool, area=area, consumable=consumable, flat=flat)
    return RateAdminForm(
        instance=instance,
        rate_history=True,
    )


def get_rate_type_choices() -> List[Tuple[str, str]]:
    # Since we want to group all tool related types, we only have a subset here
    # Return a list of (value, display value) for use in a select
    rate_type_choices = set()
    for rate_type in RateType.objects.all():
        if rate_type.is_tool_rate() and rate_type.item_specific:
            rate_type_choices.add((RateType.Type.TOOL, RateType.Type.TOOL))
        elif rate_type.is_area_rate() and rate_type.item_specific:
            rate_type_choices.add((RateType.Type.AREA, RateType.Type.AREA))
        elif rate_type.is_consumable_rate() and rate_type.item_specific:
            rate_type_choices.add((rate_type.type, rate_type.get_type_display()))
        else:
            rate_type_choices.add((rate_type.type, rate_type.get_type_display()))
    return sorted(rate_type_choices)


def get_rate_types(rate_type_choice: str) -> QuerySet[RateType]:
    rate_types = RateType.objects.none()
    if rate_type_choice:
        if rate_type_choice == RateType.Type.TOOL:
            rate_types = RateType.objects.filter(
                type__in=[
                    RateType.Type.TOOL_USAGE,
                    RateType.Type.TOOL_MISSED_RESERVATION,
                    RateType.Type.TOOL_TRAINING_INDIVIDUAL,
                    RateType.Type.TOOL_TRAINING_GROUP,
                ],
                item_specific=True,
            )
        elif rate_type_choice == RateType.Type.AREA:
            rate_types = RateType.objects.filter(
                type__in=[RateType.Type.AREA_USAGE, RateType.Type.AREA_MISSED_RESERVATION], item_specific=True
            )
        elif rate_type_choice == RateType.Type.CONSUMABLE:
            rate_types = RateType.objects.filter(type=RateType.Type.CONSUMABLE)
        else:
            rate_types = RateType.objects.filter(type=rate_type_choice)
    return rate_types


@accounting_or_manager_required
@require_GET
def delete_rate_time(request, rate_id):
    rate = get_object_or_404(Rate, pk=rate_id)
    if rate.effective_date and rate.effective_date >= datetime.today().date():
        rate.delete_with_user(request.user)
    else:
        return HttpResponseBadRequest("Past rates cannot be deleted")
    return HttpResponse()


def rate_list_permission(user):
    page_access = BillingRatesCustomization.get("rates_rate_list_page_access")

    return page_access == PageAccess.PUBLIC or (page_access == PageAccess.AUTHENTICATED and user.is_active)


@user_passes_test(rate_list_permission)
@require_GET
def rate_list(request, rate_type_choice: str = None):
    show_zero_rates = BillingRatesCustomization.get_bool("rates_rate_list_show_zero_rates")
    show_types = BillingRatesCustomization.get_list_int("rates_rate_list_page_show_types")
    show_categories = BillingRatesCustomization.get_list_int("rates_rate_list_page_show_categories")
    show_facilities = BillingRatesCustomization.get_bool("rates_rate_list_show_facilities")
    excluded_tools = BillingRatesCustomization.get_list_int("rates_rate_list_excluded_tools")

    show_tools_type = show_rate_type(show_types, RateType.Type.TOOL)
    show_areas_type = show_rate_type(show_types, RateType.Type.AREA)
    show_supplies_type = show_rate_type(show_types, RateType.Type.CONSUMABLE)
    show_other_type = show_rate_type(show_types, OTHER_RATE_TYPE)

    default_rate_type = RateType.Type.TOOL

    # determine default rate_type, no user choice
    if not show_tools_type:
        default_rate_type = RateType.Type.AREA

    if not show_tools_type and not show_areas_type:
        default_rate_type = RateType.Type.CONSUMABLE

    if not show_tools_type and not show_areas_type and not show_supplies_type:
        default_rate_type = OTHER_RATE_TYPE

    # determine default rate_type if user has selected a rate type that is no longer valid
    if not show_tools_type and rate_type_choice == RateType.Type.TOOL:
        rate_type_choice = None

    if not show_areas_type and rate_type_choice == RateType.Type.AREA:
        rate_type_choice = None

    if not show_supplies_type and rate_type_choice == RateType.Type.CONSUMABLE:
        rate_type_choice = None

    if not show_other_type and rate_type_choice == OTHER_RATE_TYPE:
        rate_type_choice = None

    if rate_type_choice not in [RateType.Type.TOOL, RateType.Type.AREA, RateType.Type.CONSUMABLE, OTHER_RATE_TYPE]:
        rate_type_choice = None

    rate_type = rate_type_choice or default_rate_type

    if rate_type == OTHER_RATE_TYPE:
        rate_types = RateType.objects.filter(item_specific=False)
        rate_order_by = []
        has_children = False
    else:
        rate_types = get_rate_types(rate_type)
        rate_order_by = [rate_type.lower(), "type", "category"]
        has_children = rate_types.count() > 1

    if show_types:
        rate_types = rate_types.filter(id__in=show_types)

    # For tool rates, don't display not visible tool rates, unless it's a child tool
    tool_filter = Q(tool__isnull=True) | (
        (Q(tool__visible=True) | Q(tool__parent_tool__isnull=False)) & ~Q(tool__id__in=excluded_tools)
    )
    all_rates = Rate.non_deleted().filter(type__in=rate_types).filter(tool_filter).order_by(*rate_order_by)

    rate_categories_exists = RateCategory.objects.exists()

    if show_categories:
        rate_categories = distinct_qs_value_list(all_rates.filter(category_id__in=show_categories), "category__name")
    else:
        rate_categories = distinct_qs_value_list(all_rates, "category__name")

    rate_categories = sorted(rate_categories)

    formatted_rates = []
    last_item = ""
    item_index = 0
    has_empty_category = not rate_categories_exists or rate_types.filter(category_specific=False).exists()

    if rate_type == OTHER_RATE_TYPE:
        for rate_type_index, group in groupby(all_rates, lambda x: x.type):
            items = list(group)
            item = items[0].get_item()
            is_first = False

            if last_item != item:
                last_item = item
                is_first = True
                item_index += 1

            type_rates = []
            has_rates = False
            if not rate_type_index.category_specific or not rate_categories_exists:
                has_rates = True
                type_rates.append(rate_display_with_details([items[0]], show_zero_rates))
            else:
                if has_empty_category:
                    type_rates.append("")
                for rate_category in rate_categories:
                    tool_rate_category = filter(lambda x: x.category.name == rate_category, items)
                    rate_display = rate_display_with_details(tool_rate_category, show_zero_rates)
                    if rate_display != "":
                        has_rates = True
                    type_rates.append(rate_display)
            if has_rates:
                formatted_rates.append(
                    RateListModel("", item, rate_type_index, is_first, item_index, has_children, type_rates)
                )

    else:
        for category_type_index, group in groupby(all_rates, lambda x: x.get_item()):
            items = list(group)
            has_children = len(list(groupby(items, lambda x: x.type))) > 1
            facility = ""
            if show_facilities:
                facility = category_type_index.core_facility

            item_index += 1
            if rate_types.count() > 1:
                is_first = True
                for item_rate_type in rate_types:
                    item_rates = []
                    has_rates = False
                    if not item_rate_type.category_specific or not rate_categories_exists:
                        has_rates = True
                        item_rates.append(rate_display_with_details(items, show_zero_rates))
                    else:
                        if has_empty_category:
                            item_rates.append("")
                        for rate_category in rate_categories:
                            tool_rate_category = filter(
                                lambda x: x.category.name == rate_category and x.type == item_rate_type, items
                            )
                            rate_display = rate_display_with_details(tool_rate_category, show_zero_rates)
                            if rate_display != "":
                                has_rates = True
                            item_rates.append(rate_display)
                    if has_rates:
                        formatted_rates.append(
                            RateListModel(
                                facility,
                                category_type_index,
                                item_rate_type,
                                is_first,
                                item_index,
                                has_children,
                                item_rates,
                            )
                        )

                    is_first = False

            else:
                items_type = items[0].type
                item_rates = []
                has_rates = False
                if not items_type.category_specific or not rate_categories_exists:
                    has_rates = True
                    item_rates.append(rate_display_with_details(items, show_zero_rates))
                else:
                    if has_empty_category:
                        item_rates.append("")
                    for rate_category in rate_categories:
                        tool_rate_category = filter(lambda x: x.category.name == rate_category, items)
                        rate_display = rate_display_with_details(tool_rate_category, show_zero_rates)
                        if rate_display != "":
                            has_rates = True
                        item_rates.append(rate_display)
                if has_rates:
                    formatted_rates.append(
                        RateListModel(facility, category_type_index, items_type, True, 0, has_children, item_rates)
                    )

    if show_facilities:
        formatted_rates.sort(key=lambda a: a.facility.name if a.facility else "")

    # boolean to toggle rate type column in table
    has_multiple_rate_types = rate_types.count() > 1

    dictionary = {
        "rate_type": rate_type,
        "formatted_rates": formatted_rates,
        "rate_categories": rate_categories,
        "has_empty_category": has_empty_category,
        "show_tools_type": show_tools_type,
        "show_areas_type": show_areas_type,
        "show_supplies_type": show_supplies_type,
        "show_other_type": show_other_type,
        "has_multiple_rate_types": has_multiple_rate_types,
        "item_label": "Item" if has_multiple_rate_types or not rate_types.exists() else f"{rate_types[0]} rates",
    }

    return render(request, "rates/rate_list.html", dictionary)


def rate_display_with_details(item_rates: Iterable[Rate], show_zero_rates):
    rate_display = ""

    current_and_future_rates = RateHistory(item_rates, datetime.today().date()).current_and_future_rates().values()
    current_and_future_rates = flatten(current_and_future_rates)
    for item_rate in current_and_future_rates:
        if item_rate.amount != 0 or show_zero_rates:
            rate_display += f"{item_rate.display_rate()}{' (' + item_rate.time.name + ')' if item_rate.time else ''}"
            if item_rate.effective_date and item_rate.effective_date > datetime.today().date():
                rate_display += f" (effective {format_datetime(item_rate.effective_date, 'SHORT_DATE_FORMAT')})"
            rate_display += "<br/>"

    return rate_display


def show_rate_type(show_rate_types, parent_rate_type):
    if not show_rate_types:
        return True

    if not parent_rate_type == OTHER_RATE_TYPE:
        rate_types = get_rate_types(parent_rate_type)
    else:
        rate_types = RateType.objects.filter(item_specific=False)
    return rate_types.filter(id__in=show_rate_types).exists()


class RateListModel:
    def __init__(self, facility, item, rate_type, is_first, item_index, has_children, item_rates):
        self.facility = facility
        self.item = item
        self.rate_type = rate_type
        self.is_first = is_first
        self.item_index = item_index
        self.has_children = has_children
        self.rates = item_rates


class PageAccess(object):
    PUBLIC = "public"
    AUTHENTICATED = "auth"



# Generated by Django 4.2.23 on 2026-01-07 14:47

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("rates", "0010_rate_service_fee"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="rate",
            options={"ordering": ["type", "category", "time", "effective_date"]},
        ),
    ]



import datetime
from collections import defaultdict
from decimal import Decimal
from typing import Dict, Iterable, List, Optional, Tuple

from NEMO_billing.exceptions import BillingException
from NEMO_billing.invoices.utilities import flatten
from NEMO_billing.rates.models import Rate, RateTime
from NEMO_billing.utilities import round_decimal_amount


class RateHistory:

    def __init__(self, rates: Iterable[Rate], as_of_date: datetime.date = None):
        self.as_of_date = as_of_date or datetime.date.today()
        self.rate_list = rates
        self.master_rates: Dict[Optional[RateTime], List[Rate]] = defaultdict(list)
        for rate in rates:
            self.master_rates[rate.time].append(rate)
        for rate_time, rate_list in self.master_rates.items():
            self.master_rates[rate_time] = sorted(
                rate_list, key=lambda r: r.effective_date or datetime.datetime.min.date()  # Use a fallback for None
            )

    def current_rates(self, as_of_date: datetime.date = None) -> Dict[Optional[RateTime], List[Rate]]:
        # Returns the current rates dictionary by time and sorted by ascending effective date
        # There can only be one current rate for each rate_time/type combination
        d = as_of_date or self.as_of_date
        current_rates_unique_dict = defaultdict(lambda: defaultdict())
        for rate_time, rate_list in self.master_rates.items():
            for individual_rate in rate_list:
                if individual_rate.effective_date is None or individual_rate.effective_date <= d:
                    current_rates_unique_dict[rate_time][individual_rate.natural_identifier()] = individual_rate
        current_rates_dict = defaultdict(list)
        for rate_time, rate_dict in current_rates_unique_dict.items():
            current_rates_dict[rate_time].extend(rate_dict.values())
        return current_rates_dict

    def current_rate(self, as_of_date: datetime.date = None) -> Dict[Optional[RateTime], Rate]:
        # Returns the current rate dictionary by time for only one type/category per rate_time
        # It won't work if the master_rates dict has more than one type/category per rate_time
        d = as_of_date or self.as_of_date
        current_rate_dict: Dict[Optional[RateTime], Rate] = {}
        # This only works because the master_rates dictionary is sorted by effective date (we only keep the most recent)
        # check that all rates only have one type/category/item, otherwise throw an error
        for rate_time, rate_list in self.master_rates.items():
            for individual_rate in rate_list:
                if individual_rate.effective_date is None or individual_rate.effective_date <= d:
                    if rate_time in current_rate_dict:
                        previous_rate_id = current_rate_dict[rate_time].natural_identifier()
                        new_rate_id = individual_rate.natural_identifier()
                        if previous_rate_id != new_rate_id:
                            raise BillingException(
                                f"The rates have multiple identifiers: {previous_rate_id} and {new_rate_id} and we cannot determine which one is current"
                            )
                    current_rate_dict[rate_time] = individual_rate
        return current_rate_dict

    def current_and_future_rates(self, as_of_date: datetime.date = None) -> Dict[Optional[RateTime], List[Rate]]:
        current_and_future_rates_dict = defaultdict(list)
        for rate_time, rate_list in self.current_rates(as_of_date).items():
            current_and_future_rates_dict[rate_time].extend(rate_list)
        for rate_time, rate_list in self.future_rates(as_of_date).items():
            current_and_future_rates_dict[rate_time].extend(rate_list)
        return current_and_future_rates_dict

    def future_rates(self, as_of_date: datetime.date = None) -> Dict[Optional[RateTime], List[Rate]]:
        # Returns the future rates dictionary by time and sorted by ascending effective date
        d = as_of_date or self.as_of_date
        future_rates_dict = defaultdict(list)
        for rate_time, rate_list in self.master_rates.items():
            for individual_rate in rate_list:
                if individual_rate.effective_date is not None and individual_rate.effective_date > d:
                    future_rates_dict[rate_time].append(individual_rate)
        return future_rates_dict

    def past_rates(self, as_of_date: datetime.date = None) -> Dict[Optional[RateTime], List[Rate]]:
        # Returns the past rates dictionary by time and sorted by descending effective date
        d = as_of_date or self.as_of_date
        current_rates = flatten(self.current_rates(as_of_date).values())
        past_rates_dict = defaultdict(list)
        for rate_time, rate_list in self.master_rates.items():
            for individual_rate in rate_list:
                if individual_rate not in current_rates and (
                    individual_rate.effective_date is None or individual_rate.effective_date <= d
                ):
                    past_rates_dict[rate_time].append(individual_rate)
        return past_rates_dict

    def all_rates(self) -> List[Rate]:
        return flatten(self.master_rates.values())


def get_rate_history(
    rate_type_id,
    rates: List[Rate] = None,
    category_id=None,
    tool_id=None,
    area_id=None,
    consumable_id=None,
    as_of_date: datetime.date = None,
) -> RateHistory:
    # if we are not given a list of rates, we get them from the database and filter directly
    if rates is None:
        matching_rates = Rate.non_deleted().filter(
            type_id=rate_type_id,
            category_id=category_id,
            tool_id=tool_id,
            area_id=area_id,
            consumable=consumable_id,
        )
    # otherwise we filter the given rates programmatically
    else:
        matching_rates = []
        for rate in rates:
            if rate.type_id == rate_type_id:
                category_match = not category_id or rate.category_id == category_id
                tool_match = not tool_id or rate.tool_id == tool_id
                area_match = not area_id or rate.area_id == area_id
                consumable_match = not consumable_id or rate.consumable_id == consumable_id
                if category_match and tool_match and area_match and consumable_match:
                    matching_rates.append(rate)

    return RateHistory(matching_rates, as_of_date)


def get_pro_rated_minimum_charge_and_service_fee(
    duration_and_rates: List[Tuple[Rate, Decimal]],
) -> Tuple[Decimal, Decimal]:
    total_duration = Decimal(0)
    total_minimum_charge = Decimal(0)
    total_service_fee = Decimal(0)
    for duration_and_rate in duration_and_rates:
        rate, duration = duration_and_rate
        total_duration += duration
        total_minimum_charge += (rate.minimum_charge or Decimal(0)) * duration
        total_service_fee += (rate.service_fee or Decimal(0)) * duration
    minimum_charge = round_decimal_amount(total_minimum_charge / total_duration) if total_duration else Decimal(0)
    service_fee = round_decimal_amount(total_service_fee / total_duration) if total_duration else Decimal(0)
    return minimum_charge, service_fee



# Generated by Django 4.2.23 on 2026-01-07 14:47

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("cap_discount", "0006_capdiscount_creation_time_and_more"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="capdiscount",
            options={"ordering": ["-creation_time", "configuration", "account", "user"]},
        ),
    ]



from django.test import TestCase

from NEMO_billing.models import CoreFacility


class TestCoreFacilities(TestCase):
    def test_create_without_external_id(self):
        """
        Tests that you can create a CoreFacility without providing an external_id.
        Check that the external_id is correctly stored as an empty string.
        """
        name_value = "Core Facility Name without External ID"

        entries_count = CoreFacility.objects.count()
        corefac = CoreFacility.objects.create(name=name_value)

        self.assertEqual(CoreFacility.objects.count(), entries_count + 1)
        self.assertIsNotNone(corefac.id)
        self.assertEqual(corefac.name, name_value)
        # For CharField, Django often converts None to an empty string on save if blank=True.
        self.assertIsNone(corefac.external_id)

    def test_create_with_external_id(self):
        """
        Tests that you can create a CoreFacility with an external_id.
        Check that the external_id is correctly stored.
        """
        name_value = "Core Facility Name with External ID"
        external_id_value = "External-ID-For-Core-Facility"

        entries_count = CoreFacility.objects.count()
        corefac = CoreFacility.objects.create(name=name_value, external_id=external_id_value)

        self.assertEqual(CoreFacility.objects.count(), entries_count + 1)
        self.assertIsNotNone(corefac.id)
        self.assertEqual(corefac.name, name_value)
        self.assertEqual(corefac.external_id, external_id_value)

    def test_core_facility_retrieval(self):
        """
        Tests that retrieving a CoreFacility matches its creation values.
        """
        name_value = "Core Facility Name to Test Retrieval"
        external_id_value = "Retrieval-External-ID"
        CoreFacility.objects.create(name=name_value, external_id=external_id_value)

        retrieved = CoreFacility.objects.get(name=name_value)

        self.assertEqual(retrieved.name, name_value)
        self.assertEqual(retrieved.external_id, external_id_value)
        self.assertIsNotNone(retrieved.id)

    def test_core_facility_retrieval_no_external_id(self):
        """
        Tests that retrieving a CoreFacility matches its creation values, with
        no external_id provided.
        """
        name_value = "Core Facility Name to Test Retrieval no External ID"
        CoreFacility.objects.create(
            name=name_value,
        )

        retrieved = CoreFacility.objects.get(name=name_value)

        self.assertEqual(retrieved.name, name_value)
        self.assertIsNone(retrieved.external_id)
        self.assertIsNotNone(retrieved.id)

    def test_core_facility_retrieval_empty_external_id(self):
        """
        Tests that retrieving a CoreFacility matches its creation values, with
        external_id as an empty string.
        """
        name_value = "Core Facility Name to Test Retrieval no External ID"
        CoreFacility.objects.create(name=name_value, external_id="")

        retrieved = CoreFacility.objects.get(name=name_value)

        self.assertEqual(retrieved.name, name_value)
        self.assertEqual(retrieved.external_id, "")
        self.assertIsNotNone(retrieved.id)



from datetime import date, datetime, time, timedelta
from http import HTTPStatus
from typing import List, Optional, Union

from NEMO.models import (
    Area,
    AreaAccessRecord,
    Consumable,
    ConsumableWithdraw,
    Project,
    Reservation,
    StaffCharge,
    Tool,
    TrainingSession,
    UsageEvent,
    User,
)
from NEMO.policy import NEMOPolicy
from NEMO.templatetags.custom_tags_and_filters import app_installed
from NEMO.utilities import format_datetime
from django.db import models
from django.db.models import Q
from django.dispatch import receiver
from django.http import HttpResponse, HttpResponseBadRequest
from django.utils.timezone import make_aware

from NEMO_billing.exceptions import BillingException, ChargeTypeNotAllowedForProjectException, HardCAPReachedException
from NEMO_billing.invoices.processors import BillableItem
from NEMO_billing.models import ProjectBillingHardCap
from NEMO_billing.utilities import get_billable_item_type_for_item, get_charges_amount_between

# Some validation for prepayment charges is tricky.
# We want to avoid using signals as they raise a hard exception, and it's challenging to catch it
# So we are using any other method unless we have no choice
# 1. Tool usage, all good we have a policy method containing the project
# 2. Staff charge, all good we can use check_billing_to_project directly
# 3. Area access record is not possible because we don't have the project
# 4. Training we don't have the project either
# 5. Consumables we don't know if those charges are from post usage or not. Either way we cannot separate at the moment
# 6. Custom charges we can use check_billing_to_project
# 7. Missed reservation cannot be dissociated from tool usage or area access
# Bonus: we are also preventing reservation on tool and areas if those are not enabled


# Decided to use signals to prevent any charge not authorized by the fund (see below)
class BillingPolicy(NEMOPolicy):
    def check_to_enable_tool(
        self, tool: Tool, operator: User, user: User, project: Project, staff_charge: bool, remote_work: bool = False
    ):
        response = super().check_to_enable_tool(tool, operator, user, project, staff_charge)
        if response.status_code != HTTPStatus.OK:
            return response
        try:
            prepayment_policy = get_prepayment_policy()
            if prepayment_policy:
                prepayment_policy.check_to_enable_tool(tool, operator, user, project, staff_charge, remote_work)
        except ChargeTypeNotAllowedForProjectException as e:
            return HttpResponseBadRequest(e.msg)
        return HttpResponse()

    def check_to_save_reservation(
        self,
        cancelled_reservation: Optional[Reservation],
        new_reservation: Reservation,
        user_creating_reservation: User,
        explicit_policy_override: bool,
    ) -> (List[str], bool):
        try:
            prepayment_policy = get_prepayment_policy()
            if prepayment_policy:
                prepayment_policy.check_to_save_reservation(
                    cancelled_reservation, new_reservation, user_creating_reservation, explicit_policy_override
                )
        except ChargeTypeNotAllowedForProjectException as e:
            return [e.msg], False
        return super().check_to_save_reservation(
            cancelled_reservation, new_reservation, user_creating_reservation, explicit_policy_override
        )

    def check_billing_to_project(
        self,
        project: Project,
        user: User,
        item: Union[Tool, Area, Consumable, StaffCharge] = None,
        charge: Union[UsageEvent, AreaAccessRecord, ConsumableWithdraw, StaffCharge, Reservation] = None,
        *args,
        **kwargs,
    ):
        super().check_billing_to_project(project, user, item, charge, *args, **kwargs)

        if isinstance(charge, Reservation):
            project_details = getattr(project, "projectbillingdetails", None)

            if project_details and project_details.expires_on:
                expiration_datetime = datetime.combine(
                    project_details.expires_on + timedelta(days=1), time.min
                ).astimezone()
                if charge.end > expiration_datetime:
                    project_expiration_exception = BillingException(
                        f'Your project "{project.name}" expires on {format_datetime(expiration_datetime)} which is before your reservation end date'
                    )
                    project_expiration_exception.project = project
                    project_expiration_exception.user = charge.user
                    raise project_expiration_exception

        prepayment_policy = get_prepayment_policy()
        if prepayment_policy:
            prepayment_policy.check_billing_to_project(project, user, item, charge)
        if hasattr(project, "projectbillinghardcap_set"):
            self.check_hard_cap_status_for_project(project, charge)

    def check_hard_cap_status_for_project(self, project: Project, charge):
        # We need to check that the new charge time is within the hard cap daterange
        # and that the type is in the list of CAP charges
        hard_caps = project.projectbillinghardcap_set.filter(enabled=True)
        if hard_caps:
            if charge and charge.get_start() and charge.get_end():
                start = charge.get_start().astimezone()
                end = charge.get_end().astimezone()
                hard_caps = hard_caps.exclude(start_date__lt=start, end_date__lt=start).exclude(
                    start_date__gt=end, end_date__gt=end
                )
            else:
                # If there is no charge let's use right now as the time
                charge_date = (charge.get_start() or charge.get_end()) if charge else datetime.now()
                charge_date = charge_date.astimezone()
                hard_caps = hard_caps.filter(Q(start_date__lte=charge_date.date()) | Q(start_date__isnull=True)).filter(
                    Q(end_date__gte=charge_date.date()) | Q(end_date__isnull=True)
                )
            for hard_cap in hard_caps:
                if get_billable_item_type_for_item(charge) in hard_cap.billable_charge_types:
                    hard_cap: ProjectBillingHardCap = hard_cap
                    start_datetime = make_aware(datetime.combine(hard_cap.start_date or date.min, time.min))
                    end_datetime = make_aware(datetime.combine(hard_cap.end_date or date.max, time.max))
                    charges, amount = get_charges_amount_between(
                        hard_cap.project,
                        hard_cap.configuration,
                        start_datetime,
                        end_datetime,
                        hard_cap.billable_charge_types,
                    )
                    if amount > hard_cap.amount:
                        raise HardCAPReachedException(hard_cap)


@receiver(models.signals.pre_save)
def auto_check_charge_type_for_projects(
    sender, instance: Union[AreaAccessRecord, UsageEvent, TrainingSession, StaffCharge], **kwargs
):
    # We don't need consumable withdrawals and reservations because we dealt with them in the policy
    if not issubclass(sender, (AreaAccessRecord, UsageEvent, TrainingSession, StaffCharge)):
        return
    if not instance.pk:
        prepayment_policy = get_prepayment_policy()
        if prepayment_policy:
            prepayment_policy.check_project_prepayment_charge(
                instance.project, BillableItem(instance, instance.project).item_type
            )


def get_prepayment_policy():
    try:
        # try to load prepayment policy if installed
        if app_installed("NEMO_billing.prepayments"):
            from NEMO_billing.prepayments.policy import PrepaymentPolicy

            return PrepaymentPolicy()
    except:
        pass
    return None



# Generated by Django 4.2.11 on 2024-07-31 22:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO", "0086_adjustmentrequest_new_quantity"),
        ("rates", "0006_rate_daily_split_multi_day_charges"),
    ]

    operations = [
        migrations.CreateModel(
            name="AreaHighestDailyRateGroup",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                (
                    "areas",
                    models.ManyToManyField(
                        help_text="Select all areas that require grouping daily, and for which only the most expensive one will be charged.",
                        to="NEMO.area",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from NEMO.models import Account
from NEMO.utilities import format_datetime

from NEMO_billing.cap_discount.models import CAPDiscountAmount
from NEMO_billing.exceptions import BillingException


class MissingCAPAmountException(BillingException):
    def __init__(self, month, account: Account):
        msg = f"Could not find prior CAP records for {month.strftime('%B %Y')} for account: '{account.name}'. Please generate invoices for that month and try again."
        super().__init__(msg)


class NotLatestInvoiceException(BillingException):
    def __init__(self, amount: CAPDiscountAmount):
        msg = "You can only void/delete the latest invoices when a CAP is associated with it."
        if amount:
            msg += (
                f" The latest amount for the {amount.cap_discount} is on {format_datetime(amount.amount_date, 'F Y')}"
            )
        super().__init__(msg)



# Generated by Django 3.2.16 on 2023-02-07 00:51

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0009_version_1_23_0"),
    ]

    operations = [
        migrations.AddField(
            model_name="invoicedetailitem",
            name="discount",
            field=models.DecimalField(blank=True, decimal_places=2, max_digits=14, null=True),
        ),
    ]



# Generated by Django 3.2.16 on 2023-01-23 20:36

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_billing", "0003_version_2_0_0"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("cap_discount", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="capdiscountconfiguration",
            name="split_by_user",
            field=models.BooleanField(
                default=True,
                help_text="Check this box to have a separate CAP amount per each user/account. If unchecked, the CAP will be per account only (all user charges count towards the same CAP)",
            ),
        ),
        migrations.AlterField(
            model_name="capdiscount",
            name="user",
            field=models.ForeignKey(
                blank=True,
                help_text="The user this CAP applies to",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AlterUniqueTogether(
            name="capdiscount",
            unique_together=set(),
        ),
        migrations.AddConstraint(
            model_name="capdiscount",
            constraint=models.UniqueConstraint(
                fields=("configuration", "account", "user"), name="cap_discount_capdiscount_unique_config_account_user"
            ),
        ),
        migrations.AddConstraint(
            model_name="capdiscount",
            constraint=models.UniqueConstraint(
                condition=models.Q(("user", None)),
                fields=("configuration", "account"),
                name="cap_discount_capdiscount_unique_config_account_null_user",
            ),
        ),
    ]



from typing import Dict

from NEMO.decorators import customization
from NEMO.exceptions import InvalidCustomizationException
from NEMO.models import Tool
from NEMO.views.customization import CustomizationBase
from django.core.exceptions import ValidationError
from django.core.validators import validate_comma_separated_integer_list

from NEMO_billing.rates.models import RateCategory, RateType


@customization(key="billing_rates", title="Billing rates")
class BillingRatesCustomization(CustomizationBase):
    variables = {
        "rates_hide_table": "",
        "rates_usage_hide_charges": "",
        "rates_hide_consumable_rates": "",
        "rates_expand_table": "",
        "rates_daily_per_account": "",
        "rates_show_all_categories": "",
        "rates_rate_list_page_access": "",
        "rates_rate_list_page_show_types": "",
        "rates_rate_list_page_show_categories": "",
        "rates_rate_list_show_zero_rates": "",
        "rates_rate_list_show_facilities": "",
        "rates_rate_list_excluded_tools": "",
    }

    def context(self) -> Dict:
        # Override to add list of rate categories and rate types
        dictionary = super().context()
        dictionary["rate_categories"] = RateCategory.objects.all()
        dictionary["rate_types"] = RateType.objects.all().order_by("-type")
        dictionary["tools"] = Tool.objects.all()
        dictionary["rate_list_excluded_tools"] = Tool.objects.filter(
            id__in=self.get_list_int("rates_rate_list_excluded_tools")
        )
        return dictionary

    def validate(self, name, value):
        if name == "rates_rate_list_page_show_categories" and value:
            validate_comma_separated_integer_list(value)
        if name == "rates_rate_list_page_show_types" and value:
            validate_comma_separated_integer_list(value)
        if name == "rates_rate_list_excluded_tools" and value:
            validate_comma_separated_integer_list(value)

    def save(self, request, element=None):
        errors = super().save(request, element)

        show_types = ",".join(request.POST.getlist("rates_rate_list_page_show_types_list", []))
        try:
            self.validate("rates_rate_list_page_show_types", show_types)
            type(self).set("rates_rate_list_page_show_types", show_types)
        except (ValidationError, InvalidCustomizationException) as e:
            errors["rates_rate_list_page_show_types"] = {"error": str(e.message or e.msg), "value": show_types}

        show_categories = ",".join(request.POST.getlist("rates_rate_list_page_show_categories_list", []))
        try:
            self.validate("rates_rate_list_page_show_categories", show_categories)
            type(self).set("rates_rate_list_page_show_categories", show_categories)
        except (ValidationError, InvalidCustomizationException) as e:
            errors["rates_rate_list_page_show_categories"] = {
                "error": str(e.message or e.msg),
                "value": show_categories,
            }

        exclude_tools = ",".join(request.POST.getlist("rate_list_excluded_tool_list", []))
        try:
            self.validate("rates_rate_list_excluded_tools", exclude_tools)
            type(self).set("rates_rate_list_excluded_tools", exclude_tools)
        except (ValidationError, InvalidCustomizationException) as e:
            errors["rates_rate_list_excluded_tools"] = {"error": str(e.message or e.msg), "value": exclude_tools}

        if not errors:
            from NEMO.rates import rate_class

            rate_class.load_rates()
        return errors



from datetime import datetime, time, timedelta, timezone
from decimal import Decimal
from logging import getLogger
from typing import Any, List, Optional, Tuple

from NEMO.constants import CHAR_FIELD_MEDIUM_LENGTH, CHAR_FIELD_SMALL_LENGTH
from NEMO.models import Area, BaseModel, Consumable, Tool, User
from NEMO.utilities import format_datetime, slugify_underscore
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.db import models
from django.db.models import Model, QuerySet
from django.urls import reverse
from django.utils.safestring import mark_safe
from mptt.fields import TreeForeignKey

from NEMO_billing.rates.model_diff import ModelDiff
from NEMO_billing.utilities import round_decimal_amount

model_logger = getLogger(__name__)


DAYS = ["Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday", "Sunday"]
MIDNIGHT = time(0, 0, 0, 0)


class ActionLog(object):
    ADD = 0
    DELETE = 1
    UPDATE = 2
    Choices = ((ADD, "Add"), (DELETE, "Delete"), (UPDATE, "Update"))


class RateCategory(BaseModel):
    name = models.CharField(
        max_length=CHAR_FIELD_MEDIUM_LENGTH, help_text="The name of this rate category", unique=True
    )

    def __str__(self):
        return self.name

    class Meta:
        verbose_name = "Rate category"
        verbose_name_plural = "Rate categories"
        ordering = ["name"]


class RateType(BaseModel):
    class Type(object):
        TOOL = "Tool"
        TOOL_USAGE = "TOOL_USAGE"
        TOOL_TRAINING_INDIVIDUAL = "TOOL_TRAINING_INDIVIDUAL"
        TOOL_TRAINING_GROUP = "TOOL_TRAINING_GROUP"
        TOOL_MISSED_RESERVATION = "TOOL_MISSED_RESERVATION"
        AREA = "Area"
        AREA_USAGE = "AREA_USAGE"
        AREA_MISSED_RESERVATION = "AREA_MISSED_RESERVATION"
        CONSUMABLE = "CONSUMABLE"
        STAFF_CHARGE = "STAFF_CHARGE"
        choices = [
            (
                TOOL,
                (
                    (TOOL_USAGE, "Tool usage"),
                    (TOOL_TRAINING_INDIVIDUAL, "Tool individual training"),
                    (TOOL_TRAINING_GROUP, "Tool group training"),
                    (TOOL_MISSED_RESERVATION, "Tool missed reservation"),
                ),
            ),
            (AREA, ((AREA_USAGE, "Area usage"), (AREA_MISSED_RESERVATION, "Area missed reservation"))),
            (CONSUMABLE, "Consumable/Supply"),
            (STAFF_CHARGE, "Staff charge"),
        ]

    type = models.CharField(max_length=CHAR_FIELD_SMALL_LENGTH, choices=Type.choices)
    category_specific = models.BooleanField(
        default=False,
        help_text="Check this box to make this rate type category specific (i.e. you will need to enter a rate for each category)",
    )
    item_specific = models.BooleanField(
        default=False,
        help_text="Check this box to make this rate type item specific (i.e. you will need to enter a rate for each item)",
    )

    def is_tool_rate(self):
        return self.type in [
            self.Type.TOOL_MISSED_RESERVATION,
            self.Type.TOOL_USAGE,
            self.Type.TOOL_TRAINING_INDIVIDUAL,
            self.Type.TOOL_TRAINING_GROUP,
        ]

    def is_area_rate(self):
        return self.type in [self.Type.AREA_USAGE, self.Type.AREA_MISSED_RESERVATION]

    def is_consumable_rate(self):
        return self.type == self.Type.CONSUMABLE

    def is_staff_charge_rate(self):
        return self.type == self.Type.STAFF_CHARGE

    def get_rate_group_type(self) -> str:
        if self.is_tool_rate() and self.item_specific:
            return RateType.Type.TOOL
        elif self.is_area_rate() and self.item_specific:
            return RateType.Type.AREA
        else:
            return self.type

    def can_have_rate_time(self):
        return self.type not in [
            RateType.Type.CONSUMABLE,
            RateType.Type.TOOL_TRAINING_GROUP,
            RateType.Type.TOOL_TRAINING_INDIVIDUAL,
        ]

    def clean(self):
        if (
            self.type
            and not self.item_specific
            and self.type in [self.Type.TOOL_USAGE, self.Type.AREA_USAGE, self.Type.CONSUMABLE]
        ):
            raise ValidationError({"item_specific": "This rate has to be item specific"})

    def validate_unique(self, exclude=None):
        super().validate_unique(exclude)
        if not exclude or exclude and "type" not in exclude and self.type:
            if RateType.objects.filter(type=self.type).exclude(id=self.pk).exists():
                raise ValidationError({"type": f"There is already a {self.get_type_display()} rate type"})

    def __str__(self):
        return self.get_type_display()


class RateTime(BaseModel):
    name = models.CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH, unique=True)

    def earliest_match(self, start_time: datetime, end_time: datetime) -> Optional[Tuple[datetime, datetime]]:
        # This function returns the earliest start (and end time) this time applies to in the daterange if any.
        # For example if this RateTime is MON-FRI 4-5PM, given a start (MON 3PM) and end (MON 8PM),
        # It will send back the earliest start & end date times (MON 4PM to MON 5PM) when this applies
        start_time_utc = start_time.astimezone(timezone.utc)
        end_time_utc = end_time.astimezone(timezone.utc)
        possible_dates: List[Tuple[datetime, datetime]] = []
        # Loop over 2 weeks, to make sure we go over all days
        for i in [0, 1]:
            datetime_control = start_time_utc + timedelta(weeks=i)
            for time_and_day in self.dailyschedule_set.all():
                start_datetime, end_datetime = time_and_day.datetime_range(datetime_control)
                # If there is overlap, add start to list
                if (start_time_utc < end_datetime) and (end_time_utc > start_datetime):
                    possible_dates.append((max(start_time_utc, start_datetime), min(end_time_utc, end_datetime)))
        return min(possible_dates) if possible_dates else (None, None)

    def overlaps(self, rate_time) -> bool:
        # Function returning whether this rate time overlaps with another
        # They overlap if any of their schedules overlap
        control_datetime = datetime.today()
        for daily_schedule in self.dailyschedule_set.all():
            for other_daily_schedule in rate_time.dailyschedule_set.all():
                if daily_schedule.overlaps(control_datetime, other_daily_schedule):
                    return True
        return False

    def __str__(self):
        return self.name


class DailySchedule(BaseModel):
    rate_time = models.ForeignKey(RateTime, on_delete=models.CASCADE)
    weekday = models.IntegerField(
        verbose_name="start day", choices=[(DAYS.index(day_name), day_name) for day_name in DAYS]
    )
    start_time = models.TimeField(blank=True, null=True, help_text="The start time")
    end_time = models.TimeField(blank=True, null=True, help_text="The end time")

    def datetime_range(self, control_datetime) -> (datetime, datetime):
        # Given a control datetime, a weekday, start and end time, returns a full datetime range
        beginning_of_the_week = control_datetime - timedelta(days=control_datetime.weekday())
        start_day = beginning_of_the_week + timedelta(days=self.weekday)
        start_datetime = datetime.combine(start_day, self.start_time or MIDNIGHT).astimezone(control_datetime.tzinfo)
        end_datetime = start_datetime + timedelta(seconds=self.duration().total_seconds())
        return start_datetime, end_datetime

    def overlaps(self, control_datetime: datetime, daily_schedule) -> bool:
        first_start, first_end = self.datetime_range(control_datetime)
        second_start, second_end = daily_schedule.datetime_range(control_datetime)
        return first_start < second_end and first_end > second_start

    def is_time_range_split(self):
        # Returns whether the time range is across 2 days, like 6pm - 6am
        # We have to treat midnight separately since 6pm - 12am is not split even though 6pm is technically after 12am
        return self.start_time and self.end_time and self.start_time >= self.end_time != MIDNIGHT

    def duration(self) -> timedelta:
        today = datetime.today()
        start_datetime = datetime.combine(today, self.start_time or MIDNIGHT)
        end_datetime = datetime.combine(today, self.end_time or MIDNIGHT)
        end_edge_cases = not self.end_time or self.end_time == MIDNIGHT or self.end_time == self.start_time
        if self.is_time_range_split() or end_edge_cases:
            end_datetime = end_datetime + timedelta(days=1)
        return end_datetime - start_datetime

    def clean(self):
        errors = {}
        if self.weekday == 6 and self.is_time_range_split():
            errors["end_time"] = ValidationError(
                "Sunday schedule has to end at Midnight at the latest (add a Monday schedule if needed)"
            )
        # Check for rates already using this time
        if self.rate_time and self.rate_time.pk:
            control_datetime = datetime.today()
            for rate in self.rate_time.rate_set.all():
                rate_filter = rate.rate_queryset_uniqueness()
                if rate_filter.exists():
                    # Check for overlapping times on previous rates
                    for other_rate_with_time in rate_filter:
                        for other_daily_schedule in other_rate_with_time.time.dailyschedule_set.all():
                            if self.overlaps(control_datetime, other_daily_schedule):
                                link = reverse("admin:rates_ratetime_change", args=[other_rate_with_time.time.id])
                                errors[NON_FIELD_ERRORS] = ValidationError(
                                    mark_safe(
                                        f'This time would now overlap with {other_rate_with_time}<br>Click <a href="{link}">here</a> to change the other rate time'
                                    )
                                )
        if errors:
            raise ValidationError(errors)

    def __str__(self):
        end_time = (
            format_datetime(self.end_time)
            if self.end_time and not self.end_time == MIDNIGHT and not self.is_time_range_split()
            else f"{next_as_loop(DAYS, self.get_weekday_display())}, {format_datetime(self.end_time or MIDNIGHT)}"
        )
        return f"{self.get_weekday_display()}, {format_datetime(self.start_time or MIDNIGHT)} to {end_time}"

    class Meta:
        ordering = ("weekday", "start_time")


class Rate(BaseModel):
    type = models.ForeignKey(RateType, on_delete=models.CASCADE)
    time = models.ForeignKey(RateTime, null=True, blank=True, on_delete=models.CASCADE)
    category = models.ForeignKey(RateCategory, null=True, blank=True, on_delete=models.CASCADE)
    tool = models.ForeignKey(Tool, null=True, blank=True, on_delete=models.CASCADE)
    area = TreeForeignKey(Area, null=True, blank=True, on_delete=models.CASCADE)
    consumable = models.ForeignKey(Consumable, null=True, blank=True, on_delete=models.CASCADE)
    amount = models.DecimalField(
        decimal_places=2, max_digits=8, help_text="The rate per hour if this isn't a flat rate"
    )
    effective_date = models.DateField(blank=True, null=True, help_text="The date when this rate becomes effective.")
    flat = models.BooleanField(default=False, help_text="Check this box to make this a flat rate (independent of time)")
    daily = models.BooleanField(default=False, help_text="Check this box to only charge once per day for this item")
    daily_split_multi_day_charges = models.BooleanField(
        default=True,
        help_text="Check this box to split charges spanning multiple days, leave unchecked to keep one long charge",
    )
    minimum_charge = models.DecimalField(
        decimal_places=2, max_digits=8, null=True, blank=True, help_text="The minimum charge for this rate"
    )
    service_fee = models.DecimalField(
        decimal_places=2,
        max_digits=8,
        null=True,
        blank=True,
        help_text="The service fee for this rate (always applies)",
    )
    deleted = models.BooleanField(default=False, help_text="Indicates that this rate was deleted")

    @classmethod
    def non_deleted(cls):
        return (
            cls.objects.filter(deleted=False)
            .select_related("type", "time", "category", "tool", "area", "consumable")
            .prefetch_related("time__dailyschedule_set")
        )

    def natural_identifier(self):
        """
        Generates a natural identifier (from type, category and item) that represents the rate (not the instance).
        Times, effective dates and attributes like flat or minimum amount are not part of the identifier.
        """
        item = self.get_item()
        if isinstance(item, Model):
            item_content_type_id = ContentType.objects.get_for_model(item).id
            item_id = item.id
        else:
            item_content_type_id = "none"
            item_id = slugify_underscore(item).lower() if item else "none"
        components = [
            f"type:{self.type_id}",  # Rate type ID
            f"category:{self.category_id or 'none'}",  # Category ID, use 'none' if not applicable
            f"item_content_type:{item_content_type_id or 'none'}",
            f"item:{item_id or 'none'}",  # Tool/Area/Consumable ID or 'staff_charge' or 'none'
        ]
        return "|".join(components)

    def get_item(self):
        if self.type.is_tool_rate():
            return self.tool
        elif self.type.is_area_rate():
            return self.area
        elif self.type.is_consumable_rate():
            return self.consumable
        elif self.type.is_staff_charge_rate():
            return "Staff charge"
        return None

    get_item.short_description = "Item"

    def clean(self):
        errors = {}
        if self.type_id:
            if self.type.is_tool_rate():
                if self.area or self.consumable:
                    errors[NON_FIELD_ERRORS] = ValidationError(
                        "You cannot select an area or a consumable for a tool rate"
                    )
                if self.type.item_specific and not self.tool:
                    errors["tool"] = ValidationError("You need to select a tool for this rate type")
            elif self.type.is_area_rate():
                if self.tool or self.consumable:
                    errors[NON_FIELD_ERRORS] = ValidationError(
                        "You cannot select a tool or a consumable for a area rate"
                    )
                if self.type.item_specific and not self.area:
                    errors["area"] = ValidationError("You need to select an area for this rate type")
            elif self.type.is_consumable_rate():
                if self.tool or self.area:
                    errors[NON_FIELD_ERRORS] = ValidationError(
                        "You cannot select a tool or an area for a consumable rate"
                    )
                if self.type.item_specific and not self.consumable:
                    errors["consumable"] = ValidationError("You need to select a consumable for this rate type")
                if not self.flat:
                    errors["flat"] = ValidationError("Consumable rates are flat rates")
            elif self.type.is_staff_charge_rate():
                if self.tool or self.consumable or self.area:
                    errors[NON_FIELD_ERRORS] = ValidationError(
                        "You cannot select a tool, area or consumable for a staff charge rate"
                    )
            if not self.type.item_specific and (self.tool or self.consumable or self.area):
                errors[NON_FIELD_ERRORS] = ValidationError(
                    "You cannot select a tool, area or consumable for a non item specific rate type"
                )
            if RateCategory.objects.exists() and self.type.category_specific and not self.category:
                errors["category"] = ValidationError("This rate type is category specific. Please select a category")
            if not self.type.category_specific and self.category:
                errors["category"] = ValidationError("The rate type you selected is not category specific")
            if self.daily:
                if self.type.type not in [RateType.Type.TOOL_USAGE, RateType.Type.AREA_USAGE]:
                    errors["daily"] = ValidationError("Only tool and area usage can be set to daily charge")
                if not self.flat:
                    errors["flat"] = ValidationError("Daily charges should be flat")
            # We cannot have different rate times for Consumable or Training rates
            if self.time:
                if not self.type.can_have_rate_time():
                    errors["time"] = ValidationError("This rate type doesn't allow different rate times")
                elif self.flat:
                    errors["time"] = ValidationError("A flat rate cannot have different schedule/times")
        if errors:
            raise ValidationError(errors)

    def rate_queryset_uniqueness(self) -> QuerySet:
        # This function returns a queryset used to check for other rates of the same type and category/item if applicable
        # Essentially a type that is category_specific but no categories exist is the same as non category specific
        category_specific = self.type.category_specific and RateCategory.objects.exists()
        rate_filter = Rate.non_deleted().filter(time__isnull=not self.time).exclude(pk=self.pk)
        if self.effective_date:
            rate_filter = rate_filter.filter(effective_date=self.effective_date)
        else:
            rate_filter = rate_filter.filter(effective_date__isnull=True)
        if not category_specific and not self.type.item_specific:
            rate_filter = rate_filter.filter(type=self.type)
        elif not category_specific and self.type.item_specific:
            rate_filter = self.rate_item_filter(rate_filter.filter(type=self.type))
        elif category_specific and self.type.item_specific:
            rate_filter = self.rate_item_filter(rate_filter.filter(type=self.type, category=self.category))
        elif category_specific and not self.type.item_specific:
            rate_filter = rate_filter.filter(type=self.type, category=self.category)
        return rate_filter

    def rate_item_filter(self, queryset: QuerySet) -> QuerySet:
        if self.type:
            if self.type.is_tool_rate():
                return queryset.filter(tool=self.tool)
            elif self.type.is_area_rate():
                return queryset.filter(area=self.area)
            elif self.type.is_consumable_rate():
                return queryset.filter(consumable=self.consumable)
        return queryset

    def validate_unique(self, exclude=None):
        super().validate_unique(exclude)
        if not exclude or exclude and "type" not in exclude and self.type:
            error_message = f"A rate of this type already exists or overlaps this one"
            rate_filter = self.rate_queryset_uniqueness()
            if rate_filter.exists():
                # If we have a time, check for overlapping times on previous rates
                if self.time:
                    for other_rate_with_time in rate_filter:
                        if self.time.overlaps(other_rate_with_time.time):
                            raise ValidationError(mark_safe(f"{error_message}: {str(other_rate_with_time)}"))
                else:
                    already_existing_rate = rate_filter.first()
                    raise ValidationError(mark_safe(f"{error_message}: {str(already_existing_rate)}"))

    # Use this method instead of regular save to allow rates audit log
    def save_with_user(self, user: User, force_insert=False, force_update=False, using=None, update_fields=None):
        try:
            rate_pre_save_log(self, user)
        except Exception as e:
            model_logger.exception(e)
        super().save(force_insert, force_update, using, update_fields)

    def delete_with_user(self, user: User):
        try:
            rate_pre_delete_log(self, user)
        except Exception as e:
            model_logger.exception(e)
        self.deleted = True
        self.save(update_fields=["deleted"])

    def is_hourly_rate(self):
        return self.type.type in [
            RateType.Type.TOOL_USAGE,
            RateType.Type.AREA_USAGE,
            RateType.Type.STAFF_CHARGE,
            RateType.Type.TOOL_TRAINING_GROUP,
            RateType.Type.TOOL_TRAINING_INDIVIDUAL,
            RateType.Type.AREA_MISSED_RESERVATION,
            RateType.Type.TOOL_MISSED_RESERVATION,
        ]

    def display_rate(self) -> str:
        return self.display_rate_value(self.amount, self.minimum_charge, self.service_fee)

    def display_rate_value(self, amount_value, minimum_charge, service_fee):
        amount = f"{amount_value:.2f}"
        if self.is_hourly_rate():
            if self.daily:
                # Display daily explicitly when we have an hourly rate set as daily
                amount = f"daily {amount}"
            elif self.flat:
                # Display flat explicitly when we have an hourly rate set as flat
                amount = f"flat {amount}"
            else:
                # Display hourly when we have an hourly rate not set as flat
                amount = f"{amount}/hr"
        minimum = f" ({minimum_charge:.2f} minimum)" if minimum_charge else ""
        service = f" +{service_fee:.2f} service fee" if service_fee else ""
        return f"{amount}{minimum}{service}"

    def calculate_amount(self, quantity: Decimal) -> Decimal:
        effective_quantity = quantity
        if self.is_hourly_rate():
            if self.flat:
                # If hourly rate is set as flat, disregard quantity
                effective_quantity = 1
            else:
                # Otherwise, divide by 60 since quantity is in minutes
                effective_quantity = quantity / Decimal(60)
        amount = effective_quantity * self.amount
        # we are handling minimum charges and service fees outside of this
        return round_decimal_amount(amount)

    def display_category_time(self):
        category = self.category.name if self.category else None
        rate_time = self.time.name if self.time else None
        cat_time = (
            [category, rate_time]
            if category and rate_time
            else [category] if category else [rate_time] if rate_time else []
        )
        return f" ({'/'.join(cat_time)})" if cat_time else ""

    def quote_display(self):
        item_name = f"{self.get_item()} " if self.get_item() else ""
        category = f" ({self.category})" if self.category else ""
        rate_time = f", {self.time}" if self.time else ""
        return f"[{self.type}] {item_name}{category}{rate_time}"

    def __str__(self):
        item_name = f"{self.get_item()} " if self.get_item() else ""
        category = f" ({self.category})" if self.category else ""
        rate_time = f", {self.time}" if self.time else ""
        effective_date = f", effective {format_datetime(self.effective_date)}" if self.effective_date else ""
        return f"{item_name}{self.type}{category}{rate_time}{effective_date}"

    class Meta:
        ordering = ["type", "category", "time", "effective_date"]


class AreaHighestDailyRateGroup(BaseModel):
    areas = models.ManyToManyField(
        Area,
        help_text="Select all areas that require grouping daily, and for which only the most expensive one will be charged.",
    )


class RateLog(BaseModel):
    content_type = models.ForeignKey(ContentType, null=True, on_delete=models.SET_NULL)
    object_id = models.PositiveIntegerField(null=True)
    content_object = GenericForeignKey("content_type", "object_id")
    action = models.IntegerField(choices=ActionLog.Choices)
    date = models.DateTimeField(auto_now_add=True)
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    details = models.TextField(null=True, blank=True)


def get_rate_content_object(rate: Rate):
    item = rate.get_item()
    return item if isinstance(item, models.Model) else None


def rate_pre_save_log(rate: Rate, user: User):
    try:
        original: Rate = Rate.objects.get(pk=rate.pk)
    except Rate.DoesNotExist:
        RateLog.objects.create(
            action=ActionLog.ADD,
            user=user,
            content_object=get_rate_content_object(rate),
            details=ModelDiff(rate).model_diff_display,
        )
    else:
        model_diff = ModelDiff(original, rate)
        if model_diff.has_changed():
            RateLog.objects.create(
                action=ActionLog.UPDATE,
                user=user,
                content_object=get_rate_content_object(original),
                details=model_diff.model_diff_display,
            )


def rate_pre_delete_log(rate: Rate, user: User):
    RateLog.objects.create(
        action=ActionLog.DELETE,
        user=user,
        content_object=get_rate_content_object(rate),
        details=ModelDiff(rate).model_diff_display,
    )


def next_as_loop(elements: List, match: Any, backwards=False) -> Any:
    # this function returns the next element after the match, looping back to the beginning if the match is the last one
    # or if reverse is True, the previous element before the match, looping back to the end if match is first element
    # For example, next_as_loop([1,2,3,4,5], 5) will return 1 and next_as_loop([1,2,3,4,5], 1, backwards) will return 5
    index_match = elements.index(match)
    if not backwards:
        return elements[0] if index_match == len(elements) - 1 else elements[index_match + 1]
    else:
        return elements[len(elements) - 1] if index_match == 0 else elements[index_match - 1]



from NEMO.exceptions import ProjectChargeException
from NEMO.utilities import format_daterange, format_datetime

from NEMO_billing.invoices.models import BillableItemType
from NEMO_billing.invoices.utilities import display_amount
from NEMO_billing.models import ProjectBillingHardCap


# General billing exception class
class BillingException(ProjectChargeException):
    def __init__(self, msg=None):
        project = getattr(self, "project", None)
        user = getattr(self, "user", None)
        super().__init__(project, user, msg)


class ChargeTypeNotAllowedForProjectException(BillingException):
    def __init__(self, project, charge_type: BillableItemType, msg=None):
        self.charge_type = charge_type
        new_msg = f"{charge_type.friendly_display_name()} charges are not allowed for project {project.name}"
        super().__init__(msg or new_msg)


class HardCAPReachedException(BillingException):
    def __init__(self, project_hard_cap: ProjectBillingHardCap, msg=None):
        self.project_hard_cap = project_hard_cap
        date_range_display = ""
        if project_hard_cap.start_date and project_hard_cap.end_date:
            date_range_display = f" during the period {format_daterange(project_hard_cap.start_date, project_hard_cap.end_date, d_format='SHORT_DATE_FORMAT')}"
        elif project_hard_cap.start_date:
            date_range_display = f" since {format_datetime(project_hard_cap.start_date, df='SHORT_DATE_FORMAT')}"
        elif project_hard_cap.end_date:
            date_range_display = f" until {format_datetime(project_hard_cap.end_date, df='SHORT_DATE_FORMAT')}"
        new_msg = f"You reached the maximum amount allowed of {display_amount(project_hard_cap.amount, project_hard_cap.configuration)} for this project{date_range_display}"
        self.project = project_hard_cap.project
        super().__init__(msg or new_msg)



from django.core.management import BaseCommand

from NEMO_billing.invoices.views.invoices import do_send_invoice_payment_reminder


class Command(BaseCommand):
    help = "Run every day to send payment reminders for unpaid invoices."

    def handle(self, *args, **options):
        do_send_invoice_payment_reminder()



# Generated by Django 3.2.19 on 2023-06-04 18:53

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("NEMO_billing", "0004_version_2_4_0"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("invoices", "0010_version_2_0_0"),
    ]

    operations = [
        migrations.AddField(
            model_name="projectbillingdetails",
            name="department",
            field=models.ForeignKey(
                blank=True,
                help_text="The project department",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="NEMO_billing.department",
            ),
        ),
        migrations.AddField(
            model_name="projectbillingdetails",
            name="institution",
            field=models.ForeignKey(
                blank=True,
                help_text="The project institution",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to="NEMO_billing.institution",
            ),
        ),
        migrations.AddField(
            model_name="projectbillingdetails",
            name="staff_host",
            field=models.ForeignKey(
                blank=True,
                help_text="The project host",
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



from __future__ import annotations

import datetime
from typing import List, Optional

from NEMO.constants import CHAR_FIELD_MEDIUM_LENGTH, CHAR_FIELD_SMALL_LENGTH
from NEMO.exceptions import ProjectChargeException
from NEMO.mixins import BillableItemMixin
from NEMO.models import (
    Area,
    BaseCategory,
    BaseModel,
    Consumable,
    Project,
    SerializationByNameModel,
    StaffCharge,
    Tool,
    User,
    validate_waive_information,
)
from django.core.exceptions import ValidationError
from django.core.validators import validate_comma_separated_integer_list
from django.db import models
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

from NEMO_billing.invoices.models import BillableItemType, InvoiceConfiguration
from NEMO_billing.templatetags.billing_tags import cap_discount_installed


class CoreFacility(BaseModel):
    name = models.CharField(
        max_length=CHAR_FIELD_MEDIUM_LENGTH, unique=True, help_text="The name of this core facility."
    )
    external_id = models.CharField(
        max_length=CHAR_FIELD_MEDIUM_LENGTH,
        null=True,
        blank=True,
        help_text="An external ID to associate with this core facility.",
    )

    def __str__(self):
        return self.name

    class Meta:
        ordering = ["name"]
        verbose_name_plural = "Core facilities"


class CoreRelationship(BaseModel):
    core_facility = models.ForeignKey(CoreFacility, on_delete=models.CASCADE)
    tool = models.OneToOneField(Tool, related_name="core_rel", blank=True, null=True, on_delete=models.CASCADE)
    area = models.OneToOneField(Area, related_name="core_rel", blank=True, null=True, on_delete=models.CASCADE)
    staff_charge = models.OneToOneField(
        StaffCharge, related_name="core_rel", blank=True, null=True, on_delete=models.CASCADE
    )
    consumable = models.OneToOneField(
        Consumable, related_name="core_rel", blank=True, null=True, on_delete=models.CASCADE
    )

    def get_item(self):
        return self.area or self.tool or self.consumable or self.staff_charge

    def __str__(self):
        return f"{self.get_item()} - {self.core_facility}"


class CustomCharge(BaseModel, BillableItemMixin):
    name = models.CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH, help_text="The name of this custom charge.")
    additional_details = models.CharField(
        max_length=CHAR_FIELD_MEDIUM_LENGTH, null=True, blank=True, help_text="Additional details for this charge."
    )
    customer = models.ForeignKey(
        User, related_name="custom_charge_customer", on_delete=models.CASCADE, help_text="The customer to charge."
    )
    creator = models.ForeignKey(
        User,
        related_name="custom_charge_creator",
        on_delete=models.CASCADE,
        help_text="The person who created this charge.",
    )
    project = models.ForeignKey(Project, on_delete=models.CASCADE, help_text="The project to bill for this charge.")
    date = models.DateTimeField(help_text="The date of the custom charge.")
    amount = models.DecimalField(
        decimal_places=2, max_digits=8, help_text="The amount of the charge. Use a negative amount for adjustments."
    )
    core_facility = models.ForeignKey(CoreFacility, null=True, blank=True, on_delete=models.SET_NULL)
    cap_eligible = models.BooleanField(default=False, help_text="Check this box to make this charge count towards CAP")
    validated = models.BooleanField(default=False)
    validated_by = models.ForeignKey(
        User, null=True, blank=True, related_name="custom_charge_validated_set", on_delete=models.CASCADE
    )
    waived = models.BooleanField(default=False)
    waived_on = models.DateTimeField(null=True, blank=True)
    waived_by = models.ForeignKey(
        User, null=True, blank=True, related_name="custom_charge_waived_set", on_delete=models.CASCADE
    )

    def clean(self):
        errors = {}
        if self.amount == 0:
            errors["amount"] = _("Please enter a positive or negative amount")
        if self.project_id and self.customer_id:
            try:
                from NEMO.policy import policy_class as policy

                policy.check_billing_to_project(self.project, self.customer, self, self)
            except ProjectChargeException as e:
                errors["project"] = e.msg
        if cap_discount_installed():
            # If this custom charge is cap eligible, check that there is actually a matching CAP configuration
            if self.cap_eligible:
                if self.customer_id and self.project_id:
                    from NEMO_billing.cap_discount.models import CAPDiscountConfiguration

                    rate_category = self.project.projectbillingdetails.category
                    from NEMO_billing.invoices.models import BillableItemType

                    cap_filter = CAPDiscountConfiguration.objects.filter(
                        rate_category=rate_category,
                        charge_types__contains=BillableItemType.CUSTOM_CHARGE.value,
                    )
                    if self.core_facility:
                        cap_filter = cap_filter.filter(core_facilities__in=[self.core_facility])
                    else:
                        cap_filter = cap_filter.filter(core_facilities__isnull=True)
                    if not cap_filter.exists():
                        errors["cap_eligible"] = _(
                            f"No CAP configuration accepting Custom charges exists for this rate category ({rate_category})"
                        )
        errors.update(validate_waive_information(self))
        if errors:
            raise ValidationError(errors)

    # To match BillableItemMixin
    def get_customer(self) -> User:
        return self.customer

    def get_operator(self) -> Optional[User]:
        return self.creator

    def get_start(self) -> Optional[datetime.datetime]:
        return None

    def get_end(self) -> Optional[datetime.datetime]:
        return self.date

    def __str__(self):
        return self.name


# Create and add a shortcut function to get core_facility from Tool, Consumable, Area or Staff Charge
def get_core_facility(self):
    # Add an exception for tool children, whose core facility is the parent's core facility
    if isinstance(self, Tool) and self.is_child_tool():
        return self.parent_tool.core_facility
    if hasattr(self, "core_rel"):
        return self.core_rel.core_facility


class Department(BaseCategory):
    pass


class InstitutionType(BaseCategory):
    pass


class Institution(SerializationByNameModel):
    name = models.CharField(
        max_length=CHAR_FIELD_MEDIUM_LENGTH, unique=True, help_text="The unique name for this institution"
    )
    institution_type = models.ForeignKey(InstitutionType, null=True, blank=True, on_delete=models.SET_NULL)
    state = models.CharField(max_length=CHAR_FIELD_SMALL_LENGTH, null=True, blank=True)
    country = models.CharField(
        max_length=2,
        choices=[
            ("AF", "Afghanistan"),
            ("AL", "Albania"),
            ("DZ", "Algeria"),
            ("AD", "Andorra"),
            ("AO", "Angola"),
            ("AI", "Anguilla"),
            ("AQ", "Antarctica"),
            ("AG", "Antigua & Barbuda"),
            ("AR", "Argentina"),
            ("AM", "Armenia"),
            ("AW", "Aruba"),
            ("AU", "Australia"),
            ("AT", "Austria"),
            ("AZ", "Azerbaijan"),
            ("BS", "Bahamas"),
            ("BH", "Bahrain"),
            ("BD", "Bangladesh"),
            ("BB", "Barbados"),
            ("BY", "Belarus"),
            ("BE", "Belgium"),
            ("BZ", "Belize"),
            ("BJ", "Benin"),
            ("BM", "Bermuda"),
            ("BT", "Bhutan"),
            ("BO", "Bolivia"),
            ("BA", "Bosnia & Herzegovina"),
            ("BW", "Botswana"),
            ("BV", "Bouvet Island"),
            ("BR", "Brazil"),
            ("GB", "Britain (UK)"),
            ("IO", "British Indian Ocean Territory"),
            ("BN", "Brunei"),
            ("BG", "Bulgaria"),
            ("BF", "Burkina Faso"),
            ("BI", "Burundi"),
            ("KH", "Cambodia"),
            ("CM", "Cameroon"),
            ("CA", "Canada"),
            ("CV", "Cape Verde"),
            ("BQ", "Caribbean NL"),
            ("KY", "Cayman Islands"),
            ("CF", "Central African Rep."),
            ("TD", "Chad"),
            ("CL", "Chile"),
            ("CN", "China"),
            ("CX", "Christmas Island"),
            ("CC", "Cocos (Keeling) Islands"),
            ("CO", "Colombia"),
            ("KM", "Comoros"),
            ("CD", "Congo (Dem. Rep.)"),
            ("CG", "Congo (Rep.)"),
            ("CK", "Cook Islands"),
            ("CR", "Costa Rica"),
            ("HR", "Croatia"),
            ("CU", "Cuba"),
            ("CW", "Curaçao"),
            ("CY", "Cyprus"),
            ("CZ", "Czech Republic"),
            ("CI", "Côte d'Ivoire"),
            ("DK", "Denmark"),
            ("DJ", "Djibouti"),
            ("DM", "Dominica"),
            ("DO", "Dominican Republic"),
            ("TL", "East Timor"),
            ("EC", "Ecuador"),
            ("EG", "Egypt"),
            ("SV", "El Salvador"),
            ("GQ", "Equatorial Guinea"),
            ("ER", "Eritrea"),
            ("EE", "Estonia"),
            ("SZ", "Eswatini (Swaziland)"),
            ("ET", "Ethiopia"),
            ("FK", "Falkland Islands"),
            ("FO", "Faroe Islands"),
            ("FJ", "Fiji"),
            ("FI", "Finland"),
            ("FR", "France"),
            ("GF", "French Guiana"),
            ("PF", "French Polynesia"),
            ("TF", "French S. Terr."),
            ("GA", "Gabon"),
            ("GM", "Gambia"),
            ("GE", "Georgia"),
            ("DE", "Germany"),
            ("GH", "Ghana"),
            ("GI", "Gibraltar"),
            ("GR", "Greece"),
            ("GL", "Greenland"),
            ("GD", "Grenada"),
            ("GP", "Guadeloupe"),
            ("GU", "Guam"),
            ("GT", "Guatemala"),
            ("GG", "Guernsey"),
            ("GN", "Guinea"),
            ("GW", "Guinea-Bissau"),
            ("GY", "Guyana"),
            ("HT", "Haiti"),
            ("HM", "Heard Island & McDonald Islands"),
            ("HN", "Honduras"),
            ("HK", "Hong Kong"),
            ("HU", "Hungary"),
            ("IS", "Iceland"),
            ("IN", "India"),
            ("ID", "Indonesia"),
            ("IR", "Iran"),
            ("IQ", "Iraq"),
            ("IE", "Ireland"),
            ("IM", "Isle of Man"),
            ("IL", "Israel"),
            ("IT", "Italy"),
            ("JM", "Jamaica"),
            ("JP", "Japan"),
            ("JE", "Jersey"),
            ("JO", "Jordan"),
            ("KZ", "Kazakhstan"),
            ("KE", "Kenya"),
            ("KI", "Kiribati"),
            ("KP", "Korea (North)"),
            ("KR", "Korea (South)"),
            ("KW", "Kuwait"),
            ("KG", "Kyrgyzstan"),
            ("LA", "Laos"),
            ("LV", "Latvia"),
            ("LB", "Lebanon"),
            ("LS", "Lesotho"),
            ("LR", "Liberia"),
            ("LY", "Libya"),
            ("LI", "Liechtenstein"),
            ("LT", "Lithuania"),
            ("LU", "Luxembourg"),
            ("MO", "Macau"),
            ("MG", "Madagascar"),
            ("MW", "Malawi"),
            ("MY", "Malaysia"),
            ("MV", "Maldives"),
            ("ML", "Mali"),
            ("MT", "Malta"),
            ("MH", "Marshall Islands"),
            ("MQ", "Martinique"),
            ("MR", "Mauritania"),
            ("MU", "Mauritius"),
            ("YT", "Mayotte"),
            ("MX", "Mexico"),
            ("FM", "Micronesia"),
            ("MD", "Moldova"),
            ("MC", "Monaco"),
            ("MN", "Mongolia"),
            ("ME", "Montenegro"),
            ("MS", "Montserrat"),
            ("MA", "Morocco"),
            ("MZ", "Mozambique"),
            ("MM", "Myanmar (Burma)"),
            ("NA", "Namibia"),
            ("NR", "Nauru"),
            ("NP", "Nepal"),
            ("NL", "Netherlands"),
            ("NC", "New Caledonia"),
            ("NZ", "New Zealand"),
            ("NI", "Nicaragua"),
            ("NE", "Niger"),
            ("NG", "Nigeria"),
            ("NU", "Niue"),
            ("NF", "Norfolk Island"),
            ("MK", "North Macedonia"),
            ("MP", "Northern Mariana Islands"),
            ("NO", "Norway"),
            ("OM", "Oman"),
            ("PK", "Pakistan"),
            ("PW", "Palau"),
            ("PS", "Palestine"),
            ("PA", "Panama"),
            ("PG", "Papua New Guinea"),
            ("PY", "Paraguay"),
            ("PE", "Peru"),
            ("PH", "Philippines"),
            ("PN", "Pitcairn"),
            ("PL", "Poland"),
            ("PT", "Portugal"),
            ("PR", "Puerto Rico"),
            ("QA", "Qatar"),
            ("RO", "Romania"),
            ("RU", "Russia"),
            ("RW", "Rwanda"),
            ("RE", "Réunion"),
            ("AS", "Samoa (American)"),
            ("WS", "Samoa (western)"),
            ("SM", "San Marino"),
            ("ST", "Sao Tome & Principe"),
            ("SA", "Saudi Arabia"),
            ("SN", "Senegal"),
            ("RS", "Serbia"),
            ("SC", "Seychelles"),
            ("SL", "Sierra Leone"),
            ("SG", "Singapore"),
            ("SK", "Slovakia"),
            ("SI", "Slovenia"),
            ("SB", "Solomon Islands"),
            ("SO", "Somalia"),
            ("ZA", "South Africa"),
            ("GS", "South Georgia & the South Sandwich Islands"),
            ("SS", "South Sudan"),
            ("ES", "Spain"),
            ("LK", "Sri Lanka"),
            ("BL", "St Barthelemy"),
            ("SH", "St Helena"),
            ("KN", "St Kitts & Nevis"),
            ("LC", "St Lucia"),
            ("SX", "St Maarten (Dutch)"),
            ("MF", "St Martin (French)"),
            ("PM", "St Pierre & Miquelon"),
            ("VC", "St Vincent"),
            ("SD", "Sudan"),
            ("SR", "Suriname"),
            ("SJ", "Svalbard & Jan Mayen"),
            ("SE", "Sweden"),
            ("CH", "Switzerland"),
            ("SY", "Syria"),
            ("TW", "Taiwan"),
            ("TJ", "Tajikistan"),
            ("TZ", "Tanzania"),
            ("TH", "Thailand"),
            ("TG", "Togo"),
            ("TK", "Tokelau"),
            ("TO", "Tonga"),
            ("TT", "Trinidad & Tobago"),
            ("TN", "Tunisia"),
            ("TR", "Turkey"),
            ("TM", "Turkmenistan"),
            ("TC", "Turks & Caicos Is"),
            ("TV", "Tuvalu"),
            ("UM", "US minor outlying islands"),
            ("UG", "Uganda"),
            ("UA", "Ukraine"),
            ("AE", "United Arab Emirates"),
            ("US", "United States"),
            ("UY", "Uruguay"),
            ("UZ", "Uzbekistan"),
            ("VU", "Vanuatu"),
            ("VA", "Vatican City"),
            ("VE", "Venezuela"),
            ("VN", "Vietnam"),
            ("VG", "Virgin Islands (UK)"),
            ("VI", "Virgin Islands (US)"),
            ("WF", "Wallis & Futuna"),
            ("EH", "Western Sahara"),
            ("YE", "Yemen"),
            ("ZM", "Zambia"),
            ("ZW", "Zimbabwe"),
            ("AX", "Åland Islands"),
        ],
    )
    zip_code = models.CharField(max_length=12, null=True, blank=True)

    def __str__(self):
        return self.name

    class Meta:
        ordering = ["name"]


class ProjectBillingHardCap(BaseModel):
    enabled = models.BooleanField(default=True)
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    configuration = models.ForeignKey(InvoiceConfiguration, null=True, blank=True, on_delete=models.CASCADE)
    start_date = models.DateField(null=True, blank=True)
    end_date = models.DateField(null=True, blank=True)
    amount = models.DecimalField(decimal_places=2, max_digits=14)
    charge_types = models.CharField(
        validators=[validate_comma_separated_integer_list],
        max_length=CHAR_FIELD_SMALL_LENGTH,
        help_text="List of charge types that will count towards this CAP",
    )

    @property
    def billable_charge_types(self) -> List[BillableItemType]:
        return [BillableItemType(int(value)) for value in self.charge_types.split(",") if value]

    def get_charge_types_display(self):
        return mark_safe(
            "<br>".join([charge_type.friendly_display_name() for charge_type in self.billable_charge_types])
        )

    def clean(self):
        if self.start_date and self.end_date and self.start_date > self.end_date:
            raise ValidationError(
                {
                    "start_date": _("Start date must be before end date"),
                }
            )
        if self.project_id:
            if not self.project.projectbillingdetails.no_tax and not self.configuration:
                raise ValidationError(
                    {
                        "configuration": _(
                            "Configuration is required for taxed projects. Select a configuration or make the project tax exempt"
                        )
                    }
                )

    class Meta:
        ordering = ["-start_date"]


setattr(Tool, "core_facility", property(get_core_facility))
setattr(Consumable, "core_facility", property(get_core_facility))
setattr(Area, "core_facility", property(get_core_facility))
setattr(StaffCharge, "core_facility", property(get_core_facility))



# Generated by Django 4.2.15 on 2024-10-05 21:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("invoices", "0013_alter_invoiceconfiguration_name_and_more"),
    ]

    operations = [
        migrations.AddField(
            model_name="invoicedetailitem",
            name="waived",
            field=models.BooleanField(default=False),
        ),
    ]



from collections import defaultdict
from datetime import timedelta

from NEMO.models import Notification, User
from NEMO.utilities import get_model_instance
from NEMO.views.customization import get_media_file_contents
from NEMO.views.notifications import delete_notification
from django.contrib.contenttypes.models import ContentType
from django.utils import timezone

from NEMO_billing.quotes.apps import NEMOQuotesConfig
from NEMO_billing.quotes.models import Quote, QuoteConfiguration

QUOTE_REVIEW_NOTIFICATION = "quote_review"
QUOTE_EMAIL_CATEGORY = NEMOQuotesConfig.plugin_id + 1


def available_configurations_for_user_to_create_or_approve(user):
    return [
        quote_configuration
        for quote_configuration in QuoteConfiguration.objects.all()
        if quote_configuration.can_user_create(user) or quote_configuration.can_user_approve(user)
    ]


def available_configurations_for_user_to_create(user):
    return [
        quote_configuration
        for quote_configuration in QuoteConfiguration.objects.all()
        if quote_configuration.can_user_create(user)
    ]


def can_access_quotes(user: User) -> bool:
    """
    A user can access quotes features if they are superuser or if they have permissions to create or approve quotes.
    """
    return user.is_active and (user.is_superuser or bool(available_configurations_for_user_to_create_or_approve(user)))


def can_create_quotes(user: User) -> bool:
    """
    A user can create quotes if they have permissions to create quotes in at least one configuration.
    """
    return bool(available_configurations_for_user_to_create(user))


def meets_send_email_prerequisites(quote: Quote) -> bool:
    quote_email_template = get_media_file_contents("quote_email.html")
    return bool(quote_email_template and quote.file and quote.file_access_token and bool(quote.get_all_recipients()))


def create_approval_request_notification(quote: Quote):
    expiration = timezone.now() + timedelta(days=30)
    user_to_notify = quote.configuration.get_all_reviewers()
    for user in user_to_notify:
        Notification.objects.update_or_create(
            user=user,
            notification_type=QUOTE_REVIEW_NOTIFICATION,
            content_type=ContentType.objects.get_for_model(quote),
            object_id=quote.id,
            defaults={"expiration": expiration},
        )


def delete_approval_request_notification(quote: Quote):
    delete_notification(QUOTE_REVIEW_NOTIFICATION, quote.id)


def get_approval_request_notifications_per_configuration(user: User):
    notifications = defaultdict(int)
    for notification in Notification.objects.filter(notification_type=QUOTE_REVIEW_NOTIFICATION, user=user):
        quote = get_model_instance(notification.content_type, notification.object_id)
        notifications[quote.configuration.id] += 1
    return notifications



from NEMO.admin import AtLeastOneRequiredInlineFormSet
from NEMO.utilities import format_datetime
from django import forms
from django.contrib.admin import ModelAdmin, TabularInline, display, register, widgets
from django.utils.safestring import mark_safe

from NEMO_billing.cap_discount.customization import CAPDiscountCustomization
from NEMO_billing.cap_discount.models import CAPDiscount, CAPDiscountAmount, CAPDiscountConfiguration, CAPDiscountTier
from NEMO_billing.invoices.models import BillableItemType
from NEMO_billing.utilities import IntMultipleChoiceField, disable_form_field


class CAPDiscountAmountInline(TabularInline):
    model = CAPDiscountAmount
    readonly_fields = ("new_charges",)
    extra = 1


class CAPDiscountTierAdminFormset(AtLeastOneRequiredInlineFormSet):
    pass


class CAPDiscountTierInline(TabularInline):
    model = CAPDiscountTier
    formset = CAPDiscountTierAdminFormset
    extra = 1
    min_num = 1

    def __init__(self, parent_model, admin_site):
        if parent_model == CAPDiscount:
            self.exclude = ["cap_discount_configuration"]
        else:
            self.exclude = ["cap_discount"]
        super().__init__(parent_model, admin_site)


class CAPDiscountConfigurationForm(forms.ModelForm):
    charge_types = IntMultipleChoiceField(
        choices=BillableItemType.choices(),
        required=True,
        widget=widgets.FilteredSelectMultiple(verbose_name="Types", is_stacked=False),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if kwargs.get("instance"):
            for field_name in ["rate_category", "split_by_user"]:
                disable_form_field(self, field_name)
        self.fields["charge_types"].initial = CAPDiscountCustomization.get("cap_billing_default_billable_types")
        if "reset_interval" in self.fields:
            self.fields["reset_interval"].initial = CAPDiscountCustomization.get("cap_billing_default_interval")
        if "reset_frequency" in self.fields:
            self.fields["reset_frequency"].initial = CAPDiscountCustomization.get("cap_billing_default_frequency")

    def _save_m2m(self):
        super()._save_m2m()
        # This is an edge case when first saving a model instance and no core_facilities are set
        # It will not trigger our signal (m2m_changed) for checking uniqueness unless we manually call clear
        if not self.instance.core_facilities.exists():
            self.instance.core_facilities.clear()

    class Meta:
        model = CAPDiscountConfiguration
        fields = "__all__"


class CAPDiscountForm(forms.ModelForm):
    charge_types = IntMultipleChoiceField(
        choices=BillableItemType.choices(),
        required=True,
        widget=widgets.FilteredSelectMultiple(verbose_name="Types", is_stacked=False),
    )

    class Meta:
        model = CAPDiscount
        fields = "__all__"


@register(CAPDiscount)
class CAPDiscountAdmin(ModelAdmin):
    form = CAPDiscountForm
    inlines = [CAPDiscountTierInline, CAPDiscountAmountInline]
    list_display = (
        "id",
        "account",
        "get_core_facilities",
        "rate_category",
        "user",
        "current_amount",
        "discount_display",
        "reset_frequency_display",
        "next_reset_display",
        "creation_time",
    )
    list_filter = ("configuration__core_facilities", "configuration__rate_category", "account", "user")
    readonly_fields = ["creation_time"]

    @display(ordering="configuration__core_facilities", description="Core facilities")
    def get_core_facilities(self, cap_discount: CAPDiscount):
        return mark_safe("<br>".join(cap_discount.configuration.core_facility_names))

    @display(ordering="configuration__rate_category", description="Rate category")
    def rate_category(self, cap_discount: CAPDiscount):
        return cap_discount.configuration.rate_category

    @display(description="Reset frequency")
    def reset_frequency_display(self, cap_discount: CAPDiscount):
        return cap_discount.get_recurrence_display() or "-"

    @display(ordering="capdiscounttier", description="Discount")
    def discount_display(self, cap_discount: CAPDiscount):
        return mark_safe("<br>".join(str(tier) for tier in cap_discount.capdiscounttier_set.all()))

    @display(description="Next reset")
    def next_reset_display(self, cap_discount: CAPDiscount):
        return format_datetime(cap_discount.next_reset(), "F Y") if cap_discount.next_reset() else "-"

    @display(description="Current amount")
    def current_amount(self, cap_discount: CAPDiscount):
        return cap_discount.latest_amount().end


@register(CAPDiscountConfiguration)
class CAPDiscountConfigurationAdmin(ModelAdmin):
    form = CAPDiscountConfigurationForm
    inlines = [CAPDiscountTierInline]
    list_display = (
        "id",
        "get_core_facilities",
        "rate_category",
        "discount_display",
        "reset_frequency_display",
        "creation_time",
    )
    list_filter = ("core_facilities", "rate_category")
    filter_horizontal = ["core_facilities"]
    readonly_fields = ["creation_time"]

    @display(description="Core facilities", ordering="core_facilities")
    def get_core_facilities(self, config: CAPDiscountConfiguration):
        return mark_safe("<br>".join(config.core_facility_names))

    @display(description="Reset frequency")
    def reset_frequency_display(self, config: CAPDiscountConfiguration):
        return config.get_recurrence_interval_display() or "-"

    @display(ordering="capdiscounttier", description="Discount")
    def discount_display(self, config: CAPDiscountConfiguration):
        return mark_safe("<br>".join(str(tier) for tier in config.capdiscounttier_set.all()))



import os
import secrets
from datetime import datetime, timedelta
from decimal import Decimal
from logging import getLogger

from NEMO.constants import CHAR_FIELD_MAXIMUM_LENGTH, CHAR_FIELD_MEDIUM_LENGTH, CHAR_FIELD_SMALL_LENGTH
from NEMO.fields import MultiEmailField, MultiRoleGroupPermissionChoiceField
from NEMO.models import BaseModel, Project, User
from NEMO.utilities import get_full_url, update_media_file_on_model_update
from django.core.exceptions import ValidationError
from django.core.files.base import ContentFile
from django.db import models
from django.dispatch import receiver
from django.template import Template, Context
from django.urls import reverse
from django.utils import timezone
from django.utils.text import slugify

from NEMO_billing.rates.models import Rate

quotes_logger = getLogger(__name__)


def get_quote_document_filename(quote, filename):
    quote_name = slugify(quote.name)
    now = datetime.now()
    generated_date = now.strftime("%Y-%m-%d_%H-%M-%S")
    year = now.strftime("%Y")
    ext = os.path.splitext(filename)[1]
    return f"quotes/{year}/{slugify(quote.quote_number)}_{quote_name}_{generated_date}{ext}"


class QuoteConfiguration(BaseModel):
    name = models.CharField(
        max_length=CHAR_FIELD_MEDIUM_LENGTH, unique=True, help_text="The name of this quote configuration"
    )
    expiration_in_days = models.PositiveIntegerField(
        help_text="The default number of days quotes are valid for once published", default=30
    )
    quote_numbering_template = models.CharField(
        default="{{ number|stringformat:'04d' }}",
        max_length=CHAR_FIELD_MEDIUM_LENGTH,
        help_text="Django template for quote numbering. Available variables: date_created, number",
    )
    current_quote_number = models.PositiveIntegerField(
        default=0,
        help_text="The current quote number used for this configuration. This is automatically incremented when a new quote is created.",
    )
    email_cc = MultiEmailField(
        null=True, blank=True, help_text="Email to cc the quote to. A comma-separated list can be used"
    )
    merchant_name = models.CharField(max_length=CHAR_FIELD_MEDIUM_LENGTH)
    merchant_details = models.TextField(
        null=True,
        blank=True,
        help_text="The merchant details to be included in the quote (address, phone number etc.)",
    )
    merchant_logo = models.ImageField(null=True, blank=True, upload_to="quote_logos")
    terms = models.TextField(null=True, blank=True, help_text="Terms and conditions to be included in the quote")

    currency = models.CharField(max_length=4, default="USD")
    currency_symbol = models.CharField(null=True, blank=True, max_length=4, default="$")

    tax = models.DecimalField(
        null=True, blank=True, decimal_places=3, max_digits=5, help_text="Tax in percent. For 20.5% enter 20.5"
    )
    tax_name = models.CharField(max_length=CHAR_FIELD_SMALL_LENGTH, null=True, blank=True, default="VAT")
    create_permissions = MultiRoleGroupPermissionChoiceField(
        groups=True,
        help_text="The roles/groups required for users to create this type of quote",
    )
    approval_permissions = MultiRoleGroupPermissionChoiceField(
        null=True,
        blank=True,
        groups=True,
        help_text="The roles/groups required for users to approve this type of quote",
    )

    @classmethod
    def get_approval_permissions_field(cls) -> MultiRoleGroupPermissionChoiceField:
        return cls._meta.get_field("approval_permissions")

    @classmethod
    def get_create_permissions_field(cls) -> MultiRoleGroupPermissionChoiceField:
        return cls._meta.get_field("create_permissions")

    def can_user_create(self, user) -> bool:
        return self.get_create_permissions_field().has_user_roles(self.create_permissions, user)

    def can_user_approve(self, user) -> bool:
        return self.get_approval_permissions_field().has_user_roles(self.approval_permissions, user)

    def get_all_reviewers(self):
        if not self.approval_permissions:
            return User.objects.none()

        user_sets = [self.get_approval_permissions_field().users_with_role(role) for role in self.approval_permissions]
        if len(user_sets) == 0:
            return User.objects.none()

        combined = user_sets[0]
        for qs in user_sets[1:]:
            combined |= qs
        return combined

    def __str__(self):
        return f"{self.name}"


class Quote(BaseModel):
    class Status(object):
        DRAFT = 0
        PENDING_APPROVAL = 1
        PUBLISHED = 2
        Choices = (
            (DRAFT, "Draft"),
            (PENDING_APPROVAL, "Pending Approval"),
            (PUBLISHED, "Published"),
        )

    name = models.CharField(
        null=False,
        blank=False,
        max_length=CHAR_FIELD_SMALL_LENGTH,
        help_text="Name of the quote",
    )
    quote_number = models.CharField(
        null=True,
        blank=True,
        max_length=CHAR_FIELD_MEDIUM_LENGTH,
        help_text="Quote number. This is automatically generated based on the quote configuration",
    )
    project = models.ForeignKey(
        Project,
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        help_text="Project associated with this quote",
    )
    configuration = models.ForeignKey(
        QuoteConfiguration,
        on_delete=models.PROTECT,
        help_text="Configuration for this quote",
    )
    users = models.ManyToManyField(User, blank=True, help_text="Users associated with this quote")
    emails = MultiEmailField(
        null=True, blank=True, help_text="Email addresses to send this quote to. A comma-separated list can be used"
    )
    expiration_date = models.DateField(null=True, blank=True, help_text="Date when this quote expires")
    creator = models.ForeignKey(
        User, on_delete=models.PROTECT, related_name="quotes_created", help_text="User who created this quote"
    )
    approved_by = models.ForeignKey(
        User,
        null=True,
        blank=True,
        on_delete=models.PROTECT,
        related_name="quotes_approved",
        help_text="User who approved or published this quote",
    )
    created_date = models.DateTimeField(auto_now_add=True, help_text="Date and time when this quote was created")
    updated_date = models.DateTimeField(auto_now=True, help_text="Date and time when this quote was last updated")
    published_date = models.DateTimeField(
        null=True, blank=True, help_text="Date and time when this quote was published"
    )
    last_emails_sent_date = models.DateTimeField(
        null=True,
        blank=True,
        help_text="Date and time when the last emails were sent for this quote",
    )
    status = models.IntegerField(
        choices=Status.Choices,
        null=False,
        blank=False,
        default=Status.DRAFT,
        help_text="Current status of the quote",
    )
    file = models.FileField(
        null=True, blank=True, max_length=CHAR_FIELD_MEDIUM_LENGTH, upload_to=get_quote_document_filename
    )
    file_access_token = models.CharField(
        null=True,
        blank=True,
        max_length=CHAR_FIELD_MAXIMUM_LENGTH,
        help_text="A random token used to access the quote file without authentication",
    )
    add_tax = models.BooleanField(
        default=False,
        help_text="Whether to add tax to the total amount of this quote",
    )

    @property
    def items(self):
        return self.quoteitem_set.all()

    @property
    def tax_amount(self):
        if self.configuration.tax and self.add_tax:
            return self.items_sum_total() * (self.configuration.tax / 100)
        return 0

    @property
    def tax_amount_display(self):
        return f"{self.configuration.currency_symbol}{self.tax_amount:.2f}  {self.configuration.currency}"

    def items_sum_total(self):
        return sum(item.total for item in self.items)

    @property
    def tax_display(self):
        if self.configuration.tax and self.add_tax:
            return f"{self.configuration.tax:.2f}%"
        return ""

    @property
    def total(self):
        return self.items_sum_total() + self.tax_amount

    @property
    def total_display(self):
        return f"{self.configuration.currency_symbol}{self.total:.2f} {self.configuration.currency}"

    @property
    def is_published(self) -> bool:
        return self.status == self.Status.PUBLISHED

    @property
    def is_pending_approval(self) -> bool:
        return self.status == self.Status.PENDING_APPROVAL

    @property
    def is_draft(self) -> bool:
        return self.status == self.Status.DRAFT

    @property
    def is_expired(self) -> bool:
        return self.expiration_date is not None and self.expiration_date < timezone.now().date()

    def can_edit_quote(self, user: User) -> bool:
        return self.creator == user

    def can_approve_quote(self, user: User) -> bool:
        return self.configuration.can_user_approve(user)

    def get_access_url(self):
        if self.file and self.file_access_token:
            url = get_full_url(reverse("public_quote_view", kwargs={"quote_id": self.id}))
            return f"{url}?token={self.file_access_token}"
        return None

    def permissions(self, user: User):
        requires_approval = bool(self.configuration.approval_permissions)
        user_can_edit = self.can_edit_quote(user)
        user_can_approve = self.can_approve_quote(user)
        can_edit = user_can_edit and self.is_draft
        can_edit_metadata = user_can_edit
        can_approve = user_can_approve and self.is_pending_approval
        can_send_emails = self.is_published and (user_can_edit or self.approved_by == user)
        return {
            "can_edit_metadata": can_edit_metadata,
            "can_approve": can_approve,
            "can_edit": can_edit,
            "requires_approval": requires_approval,
            "user_can_edit": user_can_edit,
            "can_send_emails": can_send_emails,
        }

    def publish(self, user: User):
        self.status = self.Status.PUBLISHED
        self.expiration_date = timezone.now().date() + timedelta(days=self.configuration.expiration_in_days)
        self.published_date = timezone.now()
        self.approved_by = user
        self.file = self.render_file()
        self.file_access_token = secrets.token_urlsafe()
        self.save()

    def submit_for_approval(self):
        self.status = self.Status.PENDING_APPROVAL
        self.save()

    def deny(self):
        self.status = self.Status.DRAFT
        self.save()

    def render_and_save(self):
        if self.is_published:
            self.file = self.render_file()
            self.save()

    def render_file(self):
        from NEMO_billing.quotes.renderers import quote_renderer_class

        content = quote_renderer_class.render(self)
        content.seek(0)
        content_file = ContentFile(content.read(), "quote." + quote_renderer_class.get_file_extension())
        content.close()
        return content_file

    def get_all_recipients(self):
        emails = []

        if self.users:
            for user in self.users.all():
                if user.email:
                    emails.append(user.email)

        if self.emails:
            for email in self.emails:
                emails.append(email)

        return emails

    def save(self, *args, **kwargs):
        if not self.quote_number:
            self.quote_number = self.generate_quote_number(True)
        super().save(*args, **kwargs)

    def generate_quote_number(self, update: bool = False):
        quote_number_template = self.configuration.quote_numbering_template
        current_number = self.configuration.current_quote_number
        try:
            template = Template(quote_number_template)
            date_created = self.created_date.date() if self.created_date else timezone.now().date()
            context = Context({"date_created": date_created, "number": current_number})
            quote_number = template.render(context)
            if update:
                self.configuration.current_quote_number = current_number + 1
                self.configuration.save()
            return quote_number
        except Exception:
            quotes_logger.warning("Error generating quote number", exc_info=True)
            return None

    def delete(self, *args, **kwargs):
        from NEMO_billing.quotes.utilities import delete_approval_request_notification

        delete_approval_request_notification(self)
        super().delete(*args, **kwargs)

    def clean(self):
        errors = {}

        if self.status == self.Status.PUBLISHED:
            if not self.published_date:
                errors["published_date"] = "Published quotes require a publication date."
            if not self.approved_by:
                errors["approved_by"] = "Published quotes require an approver."
            if not self.file:
                errors["file"] = "Published quotes require a document file."
            if not self.file_access_token:
                errors["file_access_token"] = "Published quotes require a file access token."
            if not self.expiration_date:
                errors["expiration_date"] = "Published quotes require an expiration date."
            elif self.expiration_date <= timezone.now().date():
                errors["expiration_date"] = "The expiration date for a published quote must be in the future."
        if self.status == self.Status.PENDING_APPROVAL:
            if self.approved_by:
                errors["approved_by"] = "Quotes waiting for approval must not have an approver."
            if self.published_date:
                errors["published_date"] = "Quotes waiting for approval must not have a publication date."
            if self.file:
                errors["file"] = "Quotes waiting for approval must not have a document file."
            if self.file_access_token:
                errors["file_access_token"] = "Quotes waiting for approval must not have a file access token."
        if self.status == self.Status.DRAFT:
            if self.approved_by:
                errors["approved_by"] = "Draft quotes must not have an approver."
            if self.published_date:
                errors["published_date"] = "Draft quotes must not have a publication date."
            if self.file:
                errors["file"] = "Draft quotes must not have a document file."
            if self.file_access_token:
                errors["file_access_token"] = "Draft quotes must not have a file access token."

        if errors:
            raise ValidationError(errors)


class QuoteItem(BaseModel):
    class AmountType(object):
        NOT_APPLICABLE = 0
        FLAT = 1
        HOURLY = 2
        DAILY = 3
        Choices = (
            (NOT_APPLICABLE, "N/A"),
            (FLAT, "Flat"),
            (HOURLY, "Hourly"),
            (DAILY, "Daily"),
        )

    quantity = models.DecimalField(
        null=False,
        blank=False,
        max_digits=8,
        default=1,
        decimal_places=0,
        help_text="Quantity of the item",
    )
    amount = models.DecimalField(
        null=False, blank=False, decimal_places=2, max_digits=8, help_text="Amount for this item"
    )
    minimum_charge = models.DecimalField(
        null=True, blank=True, decimal_places=2, max_digits=8, help_text="Minimum amount for this item"
    )
    service_fee = models.DecimalField(
        null=True, blank=True, decimal_places=2, max_digits=8, help_text="Service fee for this item"
    )
    description = models.CharField(
        null=False, blank=False, max_length=CHAR_FIELD_MEDIUM_LENGTH, help_text="Description for this item"
    )
    rate_type = models.IntegerField(
        choices=AmountType.Choices,
        null=False,
        blank=False,
        default=AmountType.NOT_APPLICABLE,
        help_text="Rate type (N/A, Flat, Hourly, Daily) for this item",
    )
    quote = models.ForeignKey(Quote, on_delete=models.CASCADE)

    @property
    def display_rate(self):
        amount = f"{self.amount:.2f}" if self.amount is not None else "N/A"
        if self.rate_type == self.AmountType.FLAT:
            amount = f"flat {amount}"
        elif self.rate_type == self.AmountType.HOURLY:
            amount = f"{amount}/hr"
        elif self.rate_type == self.AmountType.DAILY:
            amount = f"daily {amount}"
        minimum = f" ({self.minimum_charge:.2f} minimum)" if self.minimum_charge else ""
        service = f" +{self.service_fee:.2f} service fee" if self.service_fee else ""
        return f"{amount}{minimum}{service}"

    @property
    def total(self):
        amount = self.amount
        minimum_charge = self.minimum_charge or Decimal(0)
        service_fee = self.service_fee or Decimal(0)
        return max(amount * self.quantity, minimum_charge) + service_fee

    @property
    def total_display(self):
        return f"{self.quote.configuration.currency_symbol}{self.total:.2f} {self.quote.configuration.currency}"

    def clean(self):
        errors = {}

        if self.quantity <= 0:
            errors["quantity"] = "Quantity must be greater than zero."

        if self.description and self.description.strip() == "":
            errors["description"] = "Description cannot be empty."

        if self.minimum_charge is not None and self.minimum_charge < 0:
            errors["minimum_charge"] = "Minimum charge cannot be negative."

        if self.service_fee is not None and self.service_fee < 0:
            errors["service_fee"] = "Service fee cannot be negative."

        if errors:
            raise ValidationError(errors)


@receiver(models.signals.post_delete, sender=Quote)
def auto_delete_file_on_quote_delete(sender, instance: Quote, **kwargs):
    if instance.file:
        instance.file.delete(False)


@receiver(models.signals.pre_save, sender=Quote)
def auto_update_file_on_quote_change(sender, instance: Quote, **kwargs):
    return update_media_file_on_model_update(instance, "file")
